# -*- coding: utf-8 -*-
""" ECS Task definition """

from troposphere import (
    Ref, GetAtt, Tags, If
)
from troposphere.ecs import (
    PortMapping, Environment,
    LogConfiguration,
    ContainerDefinition,
    TaskDefinition
)

from ecs_composex.common import (
    add_parameters,
    KEYISSET
)
from ecs_composex.ecs.ecs_params import (
    NETWORK_MODE,
    EXEC_ROLE_T,
    TASK_ROLE_T,
    TASK_T
)
from ecs_composex.ecs import ecs_params, ecs_conditions


def import_env_variables(service):
    """
    Function to import Docker compose env variables into ECS Env Variables
    :param service: service definition
    :type service: dict

    :return: list of Environment
    :type: list<troposphere.ecs.Environment>
    """
    env_vars = []
    if KEYISSET('environment', service):
        for key in service['environment']:
            env_vars.append(
                Environment(
                    Name=key,
                    Value=service['environment'][key]
                )
            )
    return env_vars


def generate_port_mappings(network_settings):
    """
    Generates a port mapping from the Docker compose file.
    Given we are going to use AWS VPC mode, we are only considering the app port.

    :returns: mappings, list of port mappings
    :rtype: list
    """
    mappings = []
    for port in network_settings['ports']:
        mappings.append(PortMapping(
            ContainerPort=port['target'],
            HostPort=port['target']
        ))
    return mappings


def generate_container_definition(service, network_settings):
    """
    Generates the container definition
    """
    mappings = Ref('AWS::NoValue')
    env_vars = import_env_variables(service)
    mappings = generate_port_mappings(network_settings)
    container = ContainerDefinition(
        # EntryPoint=If(ENTRY_CON, Ref('AWS::NoValue'), Split(' ', Ref(params.SERVICE_ENTRYPOINT))),
        # Command=If(COMMAND_CON, Ref('AWS::NoValue'), Split('!!', Ref(params.SERVICE_COMMAND))),

        Image=Ref(ecs_params.SERVICE_IMAGE),
        Name=Ref(ecs_params.SERVICE_NAME),
        MemoryReservation=If(
            ecs_conditions.USE_FARGATE_CON_T,
            ecs_params.FARGATE_RAM,
            If(
                ecs_conditions.MEM_RES_IS_MEM_ALLOC_CON_T,
                Ref(ecs_params.MEMORY_ALLOC),
                Ref(ecs_params.MEMORY_RES)
            )
        ),
        PortMappings=mappings,
        Environment=env_vars,
        LogConfiguration=LogConfiguration(
            LogDriver='awslogs',
            Options={
                'awslogs-group': Ref(ecs_params.CLUSTER_NAME),
                'awslogs-region': Ref('AWS::Region'),
                'awslogs-stream-prefix': Ref(ecs_params.LOG_GROUP)
            }
        )
    )
    return container


def add_task_defnition(template, service_name, service, network_settings):
    """
    Function to generate and add the task definition with container definitions
    to the service template

    :param network_settings: network settings as defined in compile_network_settings
    :type network_settings: dict
    :param template: The service template to add definitions to
    :type template: troposphere.Template
    :param service_name: Name of the service
    :type service_name: str
    :param service: the service dict object
    :type service: dict
    """
    add_parameters(
        template,
        [
            ecs_params.MEMORY_ALLOC,
            ecs_params.MEMORY_RES,
            ecs_params.SERVICE_NAME,
            ecs_params.SERVICE_IMAGE,
            ecs_params.FARGATE_CPU_RAM_CONFIG,
            ecs_params.TASK_CPU_COUNT
        ]
    )
    template.add_condition(
        ecs_conditions.USE_FARGATE_CON_T,
        ecs_conditions.USE_FARGATE_CON
    )
    parameters = {
        ecs_params.SERVICE_IMAGE_T: service['image'],
        ecs_params.SERVICE_NAME_T: service_name
    }
    TaskDefinition(
        TASK_T,
        template=template,
        Cpu=If(
            ecs_conditions.USE_FARGATE_CON_T,
            ecs_params.FARGATE_CPU,
            Ref(ecs_params.TASK_CPU_COUNT)
        ),
        Memory=If(
            ecs_conditions.USE_FARGATE_CON_T,
            ecs_params.FARGATE_RAM,
            Ref(ecs_params.MEMORY_ALLOC),
        ),
        NetworkMode=NETWORK_MODE,
        Family=Ref(ecs_params.SERVICE_NAME),
        TaskRoleArn=GetAtt(template.resources[TASK_ROLE_T], 'Arn'),
        ExecutionRoleArn=GetAtt(template.resources[EXEC_ROLE_T], 'Arn'),
        ContainerDefinitions=[
            generate_container_definition(service, network_settings)
        ],
        RequiresCompatibilities=[
            'EC2',
            'FARGATE'
        ],
        Tags=Tags(
            {
                'Name': Ref(ecs_params.SERVICE_NAME),
                'Environment': Ref('AWS::StackName')
            }
        )
    )
    return parameters



# -*- coding: utf-8 -*-

"""
Main module generating the ECS Cluster template.
"""

from troposphere import Ref, If, GetAtt
from troposphere.cloudformation import Stack
from ecs_composex.compute import cluster_params, cluster_conditions
from ecs_composex.compute.hosts_template import add_hosts_resources
from ecs_composex.compute.spot_fleet import generate_spot_fleet_template, DEFAULT_SPOT_CONFIG
from ecs_composex.common import build_template, KEYISSET, LOG
from ecs_composex.common import cfn_conditions
from ecs_composex.common.cfn_params import ROOT_STACK_NAME, ROOT_STACK_NAME_T
from ecs_composex.common.templates import upload_template
from ecs_composex.vpc import vpc_params
from ecs_composex.ecs.ecs_params import CLUSTER_NAME


def add_spotfleet_stack(template, region_azs, compose_content, launch_template, **kwargs):
    """
    Function to build the spotfleet stack and add it to the Cluster parent template

    :param launch_template: the launch template
    :type launch_template: troposphere.ec2.LaunchTemplate
    :param template: parent cluster template
    :type template: troposphere.Template
    :param compose_content: docker / composex file content
    :type compose_content: dict
    :param region_azs: List of AWS Azs i.e. ['eu-west-1a', 'eu-west-1b']
    :type region_azs: list

    :returns: void
    """
    spot_config = None
    if KEYISSET('configs', compose_content):
        configs = compose_content['configs']
        if KEYISSET('spot_config', configs):
            spot_config = configs['spot_config']

    if spot_config:
        kwargs.update({'spot_config': spot_config})
    else:
        LOG.warn('No spot_config set in configs of ComposeX File. Setting to defaults')
        kwargs.update({'spot_config': DEFAULT_SPOT_CONFIG})
    fleet_template = generate_spot_fleet_template(region_azs, **kwargs)
    fleet_template_url = upload_template(
        fleet_template.to_json(),
        kwargs['BucketName'],
        'spot_fleet.json'
    )
    if not fleet_template_url:
        LOG.warn('Fleet template URL not returned. Not adding SpotFleet to Cluster stack')
        return
    template.add_resource(Stack(
        'SpotFleet',
        Condition=cluster_conditions.USE_SPOT_CON_T,
        TemplateURL=fleet_template_url,
        Parameters={
            ROOT_STACK_NAME_T: If(
                cfn_conditions.USE_STACK_NAME_CON_T,
                Ref('AWS::StackName'),
                Ref(ROOT_STACK_NAME)
            ),
            cluster_params.LAUNCH_TEMPLATE_ID_T: Ref(launch_template),
            cluster_params.LAUNCH_TEMPLATE_VersionNumber_T: GetAtt(
                launch_template, 'LatestVersionNumber'
            ),
            cluster_params.MAX_CAPACITY_T: Ref(cluster_params.MAX_CAPACITY),
            cluster_params.MIN_CAPACITY_T: Ref(cluster_params.MIN_CAPACITY),
            cluster_params.TARGET_CAPACITY_T: Ref(cluster_params.TARGET_CAPACITY)
        }
    ))


def generate_cluster_template(region_azs, compose_content=None, **kwargs):
    """Function that generates the ECS Cluster

    :param region_azs: List of AZs for hosts, i.e. ['eu-west-1', 'eu-west-b']
    :type region_azs: list
    :param compose_content: Compose dictionary to parse for services etc.
    :type compose_content: dict

    :return: ECS Cluster Template
    :rtype: troposphere.Template
    """
    template = build_template(
        'Cluster template generated by ECS Compose X',
        [
            cluster_params.CLUSTER_NAME,
            cluster_params.USE_FLEET,
            cluster_params.USE_ONDEMAND,
            cluster_params.ECS_AMI_ID,
            cluster_params.TARGET_CAPACITY,
            cluster_params.MIN_CAPACITY,
            cluster_params.MAX_CAPACITY,
            vpc_params.APP_SUBNETS,
            vpc_params.VPC_ID,
            CLUSTER_NAME
        ]
    )
    template.add_condition(
        cluster_conditions.MAX_IS_MIN_T,
        cluster_conditions.MAX_IS_MIN
    )
    template.add_condition(
        cluster_conditions.USE_SPOT_CON_T,
        cluster_conditions.USE_SPOT_CON
    )
    launch_template = add_hosts_resources(template)
    add_spotfleet_stack(
        template, region_azs,
        compose_content, launch_template, **kwargs
    )
    return template



# -*- coding: utf-8 -*-

"""
Functions to add to the Cluster template when people want to use SpotFleet for their ECS Cluster.
"""

from troposphere import (
    Ref, Sub, GetAtt, Select, If, Split
)

from troposphere.ec2 import (
    SpotFleet, SpotFleetRequestConfigData,
    LaunchTemplate,
    LaunchTemplateConfigs,
    LaunchTemplateSpecification,
    LaunchTemplateOverrides
)

from troposphere.applicationautoscaling import (
    ScalableTarget, StepAdjustment,
    StepScalingPolicyConfiguration,
    ScalingPolicy
)

from troposphere.cloudwatch import (
    Alarm,
    MetricDimension as CwMetricDimension
)
from troposphere.iam import Role
from ecs_composex.common import LOG, build_template
from ecs_composex.iam import service_role_trust_policy
from ecs_composex.vpc import vpc_params
from ecs_composex.compute import cluster_params, cluster_conditions


DEFAULT_SPOT_CONFIG = {
    'use_spot': True,
    'bid_price': 0.42,
    'spot_instance_types': {
        'm5a.xlarge': {
            'weight': 3
        },
        'm5a.2xlarge': {
            'weight': 7
        },
        'm5a.4xlarge': {
            'weight': 15
        }
    }
}


def add_fleet_role(template):
    """Function to create the IAM Role for the EC2 Spot Fleet

    :param template: source template to add the resources to
    :type template: troposphere.Template

    :returns: role
    :rtype: troposphere.iam.Role
    """
    role = Role(
        'IamFleetRole',
        template=template,
        AssumeRolePolicyDocument=service_role_trust_policy('spotfleet'),
        ManagedPolicyArns=[
            'arn:aws:iam::aws:policy/service-role/AmazonEC2SpotFleetAutoscaleRole',
            'arn:aws:iam::aws:policy/service-role/AmazonEC2SpotFleetTaggingRole'
        ]
    )
    return role


def define_overrides(region_azs, lt_id, lt_version, spot_config):
    """Function to generate Overrides for the SpotFleet

    """
    if isinstance(lt_id, LaunchTemplate):
        template_id = Ref(lt_id)
        template_version = GetAtt(lt_id, 'LatestVersionNumber')
    elif not isinstance(lt_id, Ref) or not isinstance(lt_version, Ref):
        raise TypeError(
            f"The launch template and/or version are of type", lt_id, lt_version,
            "Expected", LaunchTemplate, "Or", Ref
        )
    else:
        template_id = lt_id
        template_version = lt_version

    overrides = []
    configs = []
    LOG.debug(spot_config)
    for count, az in enumerate(region_azs):
        LOG.debug(az)
        for itype in spot_config['spot_instance_types']:
            overrides.append(
                LaunchTemplateOverrides(
                    SubnetId=Select(count, Split(',', vpc_params.APP_SUBNETS_IMPORT)),
                    WeightedCapacity=spot_config['spot_instance_types'][itype]['weight'],
                    InstanceType=itype
                )
            )
    configs.append(
        LaunchTemplateConfigs(
            LaunchTemplateSpecification=LaunchTemplateSpecification(
                LaunchTemplateId=template_id,
                Version=template_version
            ),
            Overrides=overrides
        )
    )
    return configs


def add_scaling_policies(template, spot_fleet, role):
    """Function to add Scaling to the SpotFleet

    :param template: source template to add the resources to
    :type template: troposphere.Template
    :param spot_fleet: the SpotFleet for reference
    :type spot_fleet: troposphere.ec2.SpotFleet
    :param role: IAM Role for the SpotFleet
    :type role: troposphere.iam.Role

    :returns: tuple(scale_in_policy, scale_out_policy)
    :rtype: tuple
    """
    target = ScalableTarget(
        f"SpotFleetScalingTarget{spot_fleet.title}",
        template=template,
        MaxCapacity=If(
            cluster_conditions.MAX_IS_MIN_T,
            Ref(cluster_params.MIN_CAPACITY),
            Ref(cluster_params.MAX_CAPACITY)
        ),
        MinCapacity=Ref(cluster_params.MIN_CAPACITY),
        ResourceId=Sub(f'spot-fleet-request/${{{spot_fleet.title}}}'),
        RoleARN=GetAtt(role, 'Arn'),
        ServiceNamespace='ec2',
        ScalableDimension='ec2:spot-fleet-request:TargetCapacity'
    )

    scale_in_policy = ScalingPolicy(
        f"FleetScalingIn{spot_fleet.title}",
        template=template,
        PolicyName=Sub(f'${{{spot_fleet.title}}}ScalingIn'),
        PolicyType='StepScaling',
        ScalingTargetId=Ref(target),
        StepScalingPolicyConfiguration=StepScalingPolicyConfiguration(
            AdjustmentType='ChangeInCapacity',
            Cooldown=300,
            MetricAggregationType='Average',
            StepAdjustments=[
                StepAdjustment(
                    MetricIntervalLowerBound=10,
                    ScalingAdjustment='-1'
                )
            ]
        )
    )

    scale_out_policy = ScalingPolicy(
        f"FleetScalingOut{spot_fleet.title}",
        template=template,
        PolicyName=Sub(f'${{{spot_fleet.title}}}CpuAverageScaleOut'),
        PolicyType='StepScaling',
        ScalingTargetId=Ref(target),
        StepScalingPolicyConfiguration=StepScalingPolicyConfiguration(
            AdjustmentType='ChangeInCapacity',
            Cooldown=300,
            MetricAggregationType='Average',
            StepAdjustments=[
                StepAdjustment(
                    MetricIntervalLowerBound=10,
                    ScalingAdjustment='1'
                )
            ]
        )
    )
    return (scale_in_policy, scale_out_policy)


def define_default_cw_alarms(template, spot_fleet, scaling_set):
    """
    Function to define CW alarms for the fleet. These default CW Alarms will purely rely on
    EC2 Metrics.

    :param template: Cluster Template
    :type template: troposphere.Template
    :param spot_fleet: The spot fleet the alarms are for
    :type spot_fleet: troposphere.ec2.SpotFleet
    :param scaling_set: Scale In/Out Policies triggered by alarams
    :type scaling_set: tuple

    :returns: void
    """
    Alarm(
        f"LowCpuAverageAlarm{spot_fleet.title}",
        template=template,
        ActionsEnabled=True,
        AlarmActions=[
            Ref(scaling_set[0])
        ],
        AlarmDescription=Sub(f"LOW CPU USAGE FOR ${{{spot_fleet.title}}}"),
        AlarmName=Sub(f'CPU_AVG_LOW_${{{spot_fleet.title}}}'),
        ComparisonOperator='LessThanOrEqualToThreshold',
        Dimensions=[
            CwMetricDimension(
                Name='FleetRequestId',
                Value=Ref(spot_fleet)
            )
        ],
        EvaluationPeriods=5,
        Period=60,
        Namespace='AWS/EC2Spot',
        MetricName='CPUUtilization',
        Statistic='Average',
        Threshold='25',
        Unit='Percent',
        TreatMissingData='notBreaching'
    )

    Alarm(
        f'HighCpuAverageAlarm{spot_fleet.title}',
        ActionsEnabled=True,
        AlarmActions=[
            Ref(scaling_set[1])
        ],
        AlarmDescription=Sub(f"LOW CPU USAGE FOR ${{{spot_fleet.title}}}"),
        AlarmName=Sub(f'CPU_AVG_HIGH_${{{spot_fleet.title}}}'),
        ComparisonOperator='GreaterThanOrEqualToThreshold',
        Dimensions=[
            CwMetricDimension(
                Name='FleetRequestId',
                Value=Ref(spot_fleet)
            )
        ],
        EvaluationPeriods=5,
        Period=60,
        Namespace='AWS/EC2Spot',
        MetricName='CPUUtilization',
        Statistic='Average',
        Threshold='65',
        Unit='Percent',
        TreatMissingData='notBreaching'
    )


def define_spot_fleet(template, region_azs, lt_id, lt_version, **kwargs):
    """
    Function to add a spot fleet to the cluster template

    :param template: Source template
    :type template: troposphere.Template
    :param region_azs: List of AZs to iterate onto, ie ['eu-west-1a', 'eu-west-1b']
    :type region_azs: list

    :returns: NIL
    """
    configs = define_overrides(region_azs, lt_id, lt_version, kwargs['spot_config'])
    role = add_fleet_role(template)
    fleet = SpotFleet(
        f"EcsClusterFleet",
        template=template,
        SpotFleetRequestConfigData=SpotFleetRequestConfigData(
            AllocationStrategy='diversified',
            ExcessCapacityTerminationPolicy='default',
            IamFleetRole=GetAtt(role, 'Arn'),
            InstanceInterruptionBehavior='terminate',
            TargetCapacity=Ref(cluster_params.TARGET_CAPACITY),
            Type='maintain',
            SpotPrice=str(kwargs['spot_config']['bid_price']),
            ReplaceUnhealthyInstances=True,
            LaunchTemplateConfigs=configs,
        )
    )
    scaling_set = add_scaling_policies(template, fleet, role)
    define_default_cw_alarms(template, fleet, scaling_set)


def generate_spot_fleet_template(region_azs, **kwargs):
    """
    Generates a standalone template for SpotFleet
    """
    template = build_template(
        'Template For SpotFleet As Part Of EcsCluster',
        [
            cluster_params.LAUNCH_TEMPLATE_ID,
            cluster_params.LAUNCH_TEMPLATE_VersionNumber,
            cluster_params.MIN_CAPACITY,
            cluster_params.MAX_CAPACITY,
            cluster_params.TARGET_CAPACITY
        ]
    )
    template.add_condition(
        cluster_conditions.MAX_IS_MIN_T,
        cluster_conditions.MAX_IS_MIN
    )
    lt_id = Ref(cluster_params.LAUNCH_TEMPLATE_ID)
    lt_version = Ref(cluster_params.LAUNCH_TEMPLATE_VersionNumber)
    define_spot_fleet(template, region_azs, lt_id, lt_version, **kwargs)
    return template



# -*- coding: utf-8 -*-
"""
Parameters related to the VPC settings. Used by ecs_composex.vpc and others
"""

from troposphere import Parameter, Sub, ImportValue
from ecs_composex.common.cfn_params import ROOT_STACK_NAME_T


VPC_TYPE = 'AWS::EC2::VPC::Id'
AMI_TYPE = 'AWS::EC2::Image::Id'
SUBNET_TYPE = 'AWS::EC2::Subnet::Id'
SUBNETS_TYPE = 'List<AWS::EC2::Subnet::Id>'
R53_TYPE = 'AWS::Route53::HostedZone::Id'
AZS_TYPE = 'List<AWS::EC2::AvailabilityZone::Name>'
SG_ID_TYPE = 'AWS::EC2::SecurityGroup::Id'
SG_NAME_TYPE = 'AWS::EC2::SecurityGroup::GroupName'


VPC_ID_T = 'VpcId'
VPC_ID = Parameter(
    VPC_ID_T, Type=VPC_TYPE
)

STORAGE_SUBNETS_CIDR_T = 'StorageSubnetsCidr'
STORAGE_SUBNETS_CIDR = Parameter(
    STORAGE_SUBNETS_CIDR_T,
    Type='CommaDelimitedList'
)
STORAGE_SUBNETS_T = 'StorageSubnets'
STORAGE_SUBNETS = Parameter(STORAGE_SUBNETS_T, Type=SUBNETS_TYPE)

BACKEND_SUBNETS_CIDR_T = 'BackendSubnetsCidr'
BACKEND_SUBNETS_CIDR = Parameter(
    BACKEND_SUBNETS_CIDR_T,
    Type='CommaDelimitedList'
)
BACKEND_SUBNETS_T = 'BackendSubnets'
BACKEND_SUBNETS = Parameter(BACKEND_SUBNETS_T, Type=SUBNETS_TYPE)

PUBLIC_SUBNETS_CIDR_T = 'PublicSubnetsCidr'
PUBLIC_SUBNETS_CIDR = Parameter(
    PUBLIC_SUBNETS_CIDR_T,
    Type='CommaDelimitedList'
)
PUBLIC_SUBNETS_T = 'PublicSubnets'
PUBLIC_SUBNETS = Parameter(PUBLIC_SUBNETS_T, Type=SUBNETS_TYPE)

APP_SUBNETS_CIDR_T = 'AppSubnetsCidr'
APP_SUBNETS_CIDR = Parameter(
    APP_SUBNETS_CIDR_T,
    Type='CommaDelimitedList'
)
APP_SUBNETS_T = 'AppSubnets'
APP_SUBNETS = Parameter(APP_SUBNETS_T, Type=SUBNETS_TYPE)

USE_SUB_ZONE_T = 'UseSubDnsZone'
USE_SUB_ZONE = Parameter(
    USE_SUB_ZONE_T, Type='String',
    AllowedValues=['True', 'False'],
    Default='True'
)

VPC_MAP_ID_T = 'VpcDiscoveryMapId'
VPC_MAP_ID = Parameter(
    VPC_MAP_ID_T,
    Type='String',
    Default='<none>'
)

VPC_MAP_ARN_T = 'VpcDiscoveryMapArn'
VPC_MAP_ARN = Parameter(
    VPC_MAP_ARN_T,
    Type='String',
    Default='<none>'
)

VPC_DNS_ZONE_T = 'VpcDnsZoneName'
VPC_DNS_ZONE = Parameter(
    VPC_DNS_ZONE_T,
    Type='String', Default='cluster.local'
)

VPC_ID_IMPORT = ImportValue(Sub(f"${{{ROOT_STACK_NAME_T}}}-{VPC_ID_T}"))
APP_SUBNETS_IMPORT = ImportValue(Sub(f"${{{ROOT_STACK_NAME_T}}}-{APP_SUBNETS_T}"))
PUBLIC_SUBNETS_IMPORT = ImportValue(Sub(f"${{{ROOT_STACK_NAME_T}}}-{PUBLIC_SUBNETS_T}"))
STORAGE_SUBNETS_IMPORT = ImportValue(Sub(f"${{{ROOT_STACK_NAME_T}}}-{STORAGE_SUBNETS_T}"))
NAMESPACE_ID_IMPORT = ImportValue(Sub(f"${{{ROOT_STACK_NAME_T}}}-{VPC_MAP_ID_T}"))
NAMESPACE_ARN_IMPORT = ImportValue(Sub(f"${{{ROOT_STACK_NAME_T}}}-{VPC_MAP_ARN_T}"))

APP_SUBNETS_CIDR_IMPORT = ImportValue(Sub(f"${{{ROOT_STACK_NAME_T}}}-{APP_SUBNETS_CIDR_T}"))




import logging as logthings
import re
import sys
from datetime import datetime as dt
from os import environ

import yaml
from troposphere import (
    Template, Parameter
)

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader
from ecs_composex.common import cfn_params
from ecs_composex.common import cfn_conditions


DATE = dt.utcnow().isoformat()
DATE_PREFIX = dt.utcnow().strftime('%Y/%m/%d/%H%M')
NONALPHANUM = re.compile(r'[\W]+')


def KEYISSET(x, y):
    """Macro to figure if the the dictionary contains a key and that the key is not empty

    :param x: The key to check presence in the dictionary
    :type x: str
    :param y: The dictionary to check for
    :type y: dict

    :returns: True/False
    :rtype: bool
    """
    if isinstance(y, dict) and x in y.keys() and y[x]:
        return True
    return False


def KEYPRESENT(x, y):
    """Macro to figure if the the dictionary contains a key and that the key is not empty

    :param x: The key to check presence in the dictionary
    :type x: str
    :param y: The dictionary to check for
    :type y: dict

    :returns: True/False
    :rtype: bool
    """
    if isinstance(y, dict) and x in y.keys():
        return True
    return False


def init_template(description=None):
    """Function to initialize the troposphere base template

    :param description: Description used for the CFN
    :type description: str

    :returns: template
    :rtype: Template
    """
    if description is not None:
        template = Template(description)
    else:
        template = Template(f'Template generated by ECS ComposeX')
    template.set_metadata({
        'GeneratedOn': DATE
    })
    return template


def add_parameters(template, parameters):
    """Function to add parameters to the template

    :param template: the template to add the parameters to
    :type template: troposphere.Template
    :param parameters: list of parameters to add to the template
    :type parameters: list<troposphere.Parmeter>
    """
    for param in parameters:
        if not isinstance(param, Parameter):
            raise TypeError(f'Parameter must be of type', Parameter)
        template.add_parameter(param)


def add_defaults(template):
    """Function to CFN parameters and conditions to the template whhich are used
    across ECS ComposeX

    :param template: source template to add the params and conditions to
    :type template: Template
    """
    template.add_parameter(cfn_params.USE_SSM_EXPORTS)
    template.add_parameter(cfn_params.USE_CFN_EXPORTS)
    template.add_parameter(cfn_params.ROOT_STACK_NAME)

    template.add_condition(
        cfn_conditions.USE_SSM_EXPORTS_T,
        cfn_conditions.USE_SSM_EXPORTS
    )
    template.add_condition(
        cfn_conditions.USE_CFN_EXPORTS_T,
        cfn_conditions.USE_CFN_EXPORTS
    )
    template.add_condition(
        cfn_conditions.NOT_USE_CFN_EXPORTS_T,
        cfn_conditions.NOT_USE_CFN_EXPORTS
    )
    template.add_condition(
        cfn_conditions.USE_CFN_AND_SSM_EXPORTS_T,
        cfn_conditions.USE_CFN_AND_SSM_EXPORTS
    )
    template.add_condition(
        cfn_conditions.USE_STACK_NAME_CON_T,
        cfn_conditions.USE_STACK_NAME_CON
    )
    template.add_condition(
        cfn_conditions.USE_SSM_ONLY_T,
        cfn_conditions.USE_SSM_ONLY
    )


def build_template(description=None, *parameters):
    """Entry point function to creating the template for ECS ComposeX resources

    :param description: Optional custom description for the CFN template
    :type description: str, optional

    :param parameters: List of optional parameters to add to the template.
    :type parameters: List<troposphere.Parameters>, optional

    :returns template: the troposphere template
    :rtype: Template
    """
    template = init_template(description)
    if parameters:
        add_parameters(template, *parameters)
    add_defaults(template)
    return template


def validate_resource_title(resource_name, resource_type=None):
    """Function to validate the key for the resource is valid

    :param resource_name: Name of the resource to evaluate
    :type resource_name: str
    :param resource_type: category of the resource, optional
    :type resource_type: str

    :returns: True/False
    :rtype: bool
    """
    if NONALPHANUM.findall(resource_name):
        raise ValueError(
            f"The resource {resource_name} in {resource_type} "
            "section contains non alphanumerical characters",
            NONALPHANUM.findall(resource_name)
        )
    return True


def validate_input(compose_content, res_key):
    """Function to validate the resources names in ComposeX File
    for a given resource key

    :param compose_content: the docker/ComposeX content
    :type compose_content: dict
    :param res_key: key of the category in docker compose to look for
    :type res_key: str

    :return: True/False if all keys are valid
    :rtype: bool
    """
    names = compose_content[res_key]
    for name in names:
        validate_resource_title(name, res_key)
    return True


def validate_kwargs(required_keys, kwargs, caller=None):
    """Function to ensure minimum keys in kwargs are present.

    :param required_keys: the list of keys that have to be present
    :type required_keys: list
    :param kwargs: the arguments to verify
    :type kwargs: dict or set
    :param caller: optional argument to help understand what's missing
    :type caller: str

    :return: True/False if all keys are valid
    :rtype: bool

    :raises: KeyError if key is missing from kwargs
    """
    for required_key in required_keys:
        if not KEYISSET(required_key, kwargs):
            raise KeyError(required_key, 'is required by module', caller)
    return True


def setup_logging():
    """Function to setup logging for ECS ComposeX.
    In case this is used in a Lambda function, removes the AWS Lambda default log handler

    :returns: the_logger
    :rtype: Logger
    """
    level = environ.get('LOGLEVEL')

    if level is not None and isinstance(level, str):
        print('SETTING TO', level.upper())
        logthings.basicConfig(level=level.upper())
    else:
        logthings.basicConfig(level='INFO')

    root_logger = logthings.getLogger()
    for h in root_logger.handlers:
        root_logger.removeHandler(h)
    the_logger = logthings.getLogger("EcsComposeX")

    if not the_logger.handlers:
        formatter = logthings.Formatter(
            "[%(levelname)s] %(asctime)s, %(funcName)s, %(message)s", "%Y-%m-%d %H:%M:%S"
        )
        handler = logthings.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)
        the_logger.addHandler(handler)

    return the_logger


def load_composex_file(file_path):
    """File to load and read the docker compose file

    :param file_path: path to the docker compose file
    :type file_path: str

    :return: content of the docker file
    :rtype: dict
    """
    with open(file_path, 'r') as composex_fd:
        return yaml.load(composex_fd.read(), Loader=Loader)


LOG = setup_logging()



# -*- coding: utf-8 -*-
"""
Functions to format CFN template Outputs
"""

from troposphere import Output, Export, Sub, If
from ecs_composex.common.cfn_params import ROOT_STACK_NAME_T
from ecs_composex.common.cfn_conditions import USE_STACK_NAME_CON_T


def cfn_resource_type(object_name, strip=True):
    """Returns the Resource Type property removing the AWS:: prefix

    :returns: res_type
    :rtype: str
    """
    res_type = object_name.resource_type.replace(':', '')
    if strip:
        return res_type.replace('AWS', '')
    return res_type


def formatted_outputs(comments, export=False, prefix=None, use_root_stack=False):
    """Function to format the outputs easily and add exports based on a prefix

    :param comments: List of KeyPair values representing the output
    :type comments: list
    :param export: Whether or not this output should export to CFN Exports. Default: False
    :type export: bool
    :param prefix: prefix for the cfn exports. Must enable uniqueness of the output in CFN
    :type prefix: str
    :param use_root_stack: Whether this output for prefix should use the RootStackName as prefix
    :type use_root_stack: bool

    :return: outputs
    :rtype: list
    """
    outputs = []
    if isinstance(comments, list):
        for comment in comments:
            if isinstance(comment, dict):
                keys = list(comment.keys())
                args = {
                    'title': keys[0],
                    'Value': comment[keys[0]]
                }
                if export:
                    if prefix is None:
                        args['Export'] = Export(
                            If(
                                USE_STACK_NAME_CON_T,
                                Sub(f'${{AWS::StackName}}-{keys[0]}'),
                                Sub(f"${{{ROOT_STACK_NAME_T}}}-{keys[0]}")
                            )
                        )
                    elif use_root_stack:
                        Export(Sub(f'${{{ROOT_STACK_NAME_T}}}-{keys[0]}'))
                    elif isinstance(prefix, str):
                        args['Export'] = Export(Sub(f'{prefix}-{keys[0]}'))
                outputs.append(Output(**args))
    return outputs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `ecs_composex` package."""

import json
from os import environ
from ecs_composex.common.templates import (
    check_bucket, upload_template
)

try:
    BUCKET = environ['KNOWN_BUCKET']
except KeyError:
    BUCKET = 'lambda-dev-eu-west-1'


def test_s3_bucket():
    """
    Test s3 bucket
    """
    assert check_bucket(BUCKET)


def test_s3_upload():
    """
    Tests s3 upload function
    """
    upload_template(
        json.dumps({'Test': True}),
        BUCKET,
        'test.json',
        validate=False
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Console script for ecs_composex.vpc"""

import sys
import argparse
from boto3 import session

from ecs_composex.common.aws import CURATED_AZS
from ecs_composex.vpc import create_vpc_template


def main():
    """Console script for ecs_composex."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--vpc-cidr', required=False, default='192.168.36.0/22', dest='VpcCidr',
        help="Specify the VPC CIDR"
    )
    parser.add_argument(
        '--region', required=False, default=session.Session().region_name,
        dest='AwsRegion',
        help="Specify the region you want to build for"
        "default use default region from config or environment vars"
    )
    parser.add_argument(
        '--az', dest='AwsAzs', action='append', required=False, default=CURATED_AZS,
        help="List AZs you want to deploy to specifically within the region"
    )
    parser.add_argument(
        '-o', '--output-file', required=True,
        help="Output file for the template body"
    )
    parser.add_argument(
        '--single-nat', dest='SingleNat', action='store_true',
        help="Whether you want a single NAT for your application subnets or not. Not recommended for production"
    )
    parser.add_argument('_', nargs='*')
    args = parser.parse_args()

    template = create_vpc_template(**vars(args))
    if args.output_file:
        with open(args.output_file, 'w') as tpl_fd:
            if args.output_file.endswith('.yml') or args.output_file.endswith('.yaml'):
                tpl_fd.write(template.to_yaml())
            else:
                tpl_fd.write(template.to_json())
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



# -*- coding: utf-8 -*-
""" IAM Building block for ECS """


from troposphere import Sub, Ref
from troposphere.iam import Role, PolicyType

from ecs_composex.common import LOG
from ecs_composex.ecs.ecs_params import SERVICE_NAME_T, CLUSTER_NAME_T, EXEC_ROLE_T, TASK_ROLE_T, TASK_T
from ecs_composex.ecs_composex import generate_x_resource_configs
from ecs_composex.iam import service_role_trust_policy


def add_service_roles(template):
    """
    Function to create the IAM roles for the ECS task

    :param template: service template to add the resources to
    :type template: troposphere.Template
    """
    execution_role = Role(
        EXEC_ROLE_T,
        template=template,
        AssumeRolePolicyDocument=service_role_trust_policy('ecs-tasks'),
        Description=Sub(f"Execution role for ${{{SERVICE_NAME_T}}} in ${{{CLUSTER_NAME_T}}}")
    )
    PolicyType(
        f'{EXEC_ROLE_T}Policy',
        template=template,
        PolicyName=Sub(f'EcsExecRole'),
        PolicyDocument={
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "AllowsForEcrPullFromEcsAgent",
                    "Effect": "Allow",
                    "Action": [
                        "ecr:GetAuthorizationToken",
                        "ecr:BatchCheckLayerAvailability",
                        "ecr:GetDownloadUrlForLayer",
                        "ecr:GetRepositoryPolicy",
                        "ecr:DescribeRepositories",
                        "ecr:ListImages",
                        "ecr:DescribeImages",
                        "ecr:BatchGetImage"
                    ],
                    "Resource": ['*']
                },
                {
                    "Sid": "AllowEcsAgentOrientedTasks",
                    "Effect": "Allow",
                    "Action": [
                        "ecs:DiscoverPollEndpoint",
                        "ecs:Poll",
                        "ecs:Submit*",
                    ],
                    "Resource": ["*"]
                },
                {
                    "Sid": "AllowCloudWatchLoggingToSpecificLogGroup",
                    "Effect": "Allow",
                    "Action": [
                        "logs:CreateLogStream",
                        "logs:PutLogEvents"
                    ],
                    "Resource": [
                        Sub(
                            'arn:${AWS::Partition}:logs:${AWS::Region}:${AWS::AccountId}:log-group:'
                            f'${{{CLUSTER_NAME_T}}}:*'
                        )
                    ]
                },
                {
                    "Sid": "AllowsEcsAgentToPerformActionsForMicroservice",
                    "Effect": "Allow",
                    "Action": [
                        "ec2:AttachNetworkInterface",
                        "ec2:CreateNetworkInterface",
                        "ec2:CreateNetworkInterfacePermission",
                        "ec2:DeleteNetworkInterface",
                        "ec2:DeleteNetworkInterfacePermission",
                        "ec2:Describe*",
                        "ec2:DetachNetworkInterface",
                        "elasticloadbalancing:DeregisterInstancesFromLoadBalancer",
                        "elasticloadbalancing:DeregisterTargets",
                        "elasticloadbalancing:Describe*",
                        "elasticloadbalancing:RegisterInstancesWithLoadBalancer",
                        "elasticloadbalancing:RegisterTargets",
                    ],
                    "Resource": ["*"]
                }
            ]
        },
        Roles=[Ref(execution_role)]
    )
    Role(
        TASK_ROLE_T,
        template=template,
        AssumeRolePolicyDocument=service_role_trust_policy('ecs-tasks'),
        Description=Sub(f"TaskRole - ${{{SERVICE_NAME_T}}} in ${{{CLUSTER_NAME_T}}}"),
        ManagedPolicyArns=[],
        Policies=[]
    )


def define_service_containers(service_template):
    """Function to set the containers list from the service_task definition object

    :param service_template: the task definition
    :type service_template: troposphere.Template

    :return: list of containers
    :rtype: list
    """
    service_task = None
    if TASK_T in service_template.resources:
        service_task = service_template.resources[TASK_T]
    try:
        if service_task:
            containers = getattr(service_task, 'ContainerDefinitions')
        else:
            containers = []
    except AttributeError:
        raise ValueError('Service Task definition defined but no ContainerDefinitions found')
    return containers


def set_resource_type_settings(service_template, service_name, resources_permissions, env_vars, containers):
    """Function to add the resource type settings for a service if applicable

    :param service_template: the service template
    :type service_template: troposphere.Template
    :param service_name: name of the service as defined in the docker compose file
    :type service_name: str
    :param resources_permissions: resource permissions for a given resource type
    :type resources_permissions: dict
    :param env_vars: environment variables to get from the extra resources to add to the container environment
    :type env_vars: dict
    :param containers: list of containers from the task definition
    :type containers: list<troposphere.ecs.ContainerDefinition>
   """
    task_role = service_template.resources[TASK_ROLE_T]
    for resource_name in resources_permissions:
        res_vars = env_vars[resource_name]
        for permission_type in resources_permissions[resource_name]:
            permission = resources_permissions[resource_name][permission_type]
            if service_name in permission['Services']:
                task_role.Policies.append(permission['Policy'])
                for container in containers:
                    environment = getattr(container, 'Environment')
                    environment += res_vars
                    LOG.debug(environment)


def assign_x_resources_to_service(compose_content, service_name, service_tpl, **kwargs):
    """
    Parses each X component and if the service is listed there, assigns the policy to
    the service task role

    :param compose_content: docker ComposeX file content
    :type compose_content: dict
    :param service_name: name of the service
    :type service_name: str
    :param service_tpl: service template to add the resources to
    :type service_tpl: troposphere.Template
    :param kwargs: optional arguments
    :type kwargs: dict
    """
    x_resources_configs = generate_x_resource_configs(
        compose_content, **kwargs
    )
    containers = define_service_containers(service_tpl)

    for resource_type in x_resources_configs:
        if not (resource_type == 'x-rds' or resource_type == 'x-cluster'):
            resources_perms = x_resources_configs[resource_type]['permissions']
            env_vars = x_resources_configs[resource_type]['envvars']
            set_resource_type_settings(service_tpl, service_name, resources_perms, env_vars, containers)



# -*- coding: utf-8 -*-
"""Common variables fetched from AWS."""

import boto3


def get_region_azs(region=None, session=None, client=None):
    """Function to return the AZ from a given region. Uses default region for this

    :param region: override region_name for boto3.session.Session
    :type region: str
    :param session: override boto3.session.Session for the next client request
    :type session: boto3.session.Session()
    :param client: boto3 client
    :type client: boto3.client()

    :return: list of AZs in the given region
    :rtype: list
    """
    if session is None:
        if region is None:
            session = boto3.session.Session()
        elif isinstance(region, str):
            session = boto3.session.Session(region_name=region)
    if client is None:
        return session.client('ec2').describe_availability_zones()['AvailabilityZones']
    return client.describe_availability_zones()['AvailabilityZones']


def get_curated_azs(region=None, session=None, client=None):
    """Function to return curated list of AZs

    :param region: override region_name for boto3.session.Session
    :type region: str
    :param session: override boto3.session.Session for the next client request
    :type session: boto3.session.Session
    :param client: boto3 client to make the API call
    :type client: boto3.client

    :return: list of AZs from AWS
    :rtype: list
    """
    azs = get_region_azs(region, session, client)
    return [az['ZoneName'] for az in azs]


def get_account_id(session=None, client=None):
    """
    Function to get the current session account ID

    :param session: boto3 session to override API calls
    :type session: boto3.session.Session
    :param client: boto3 client to make API calls
    :type client: boto3.client

    :return: list of AZs
    :rtype: list
    """
    if client is not None:
        return client.get_caller_identity()['Account']
    elif client is None and session is not None:
        return session.client('sts').get_caller_identity()['Account']
    elif client is None and session is None:
        return boto3.client('sts').get_caller_identity()['Account']


CURATED_AZS = get_curated_azs()
ACCOUNT_ID = get_account_id()
BUCKET_NAME = f"cfn-templates-{ACCOUNT_ID[:6]}"



# -*- coding: utf-8 -*-

"""Generates the individual SQS Queues templates."""

import os

from troposphere import (
    Tags, Sub, GetAtt,
    ImportValue, If, Ref
)
from troposphere.cloudformation import Stack
from troposphere.sqs import (
    Queue, RedrivePolicy
)
from troposphere.ssm import (
    Parameter as SsmParameter
)

from ecs_composex.common import LOG
from ecs_composex.common import (
    build_template,
    cfn_params, KEYISSET, KEYPRESENT,
)
from ecs_composex.common import validate_kwargs
from ecs_composex.common.cfn_conditions import (
    USE_SSM_ONLY_T,
    USE_SSM_EXPORTS_T,
    USE_STACK_NAME_CON_T
)
from ecs_composex.common.cfn_params import (
    ROOT_STACK_NAME,
    ROOT_STACK_NAME_T
)
from ecs_composex.common.outputs import formatted_outputs
from ecs_composex.common.templates import (
    upload_template
)
from ecs_composex.sqs.sqs_params import (
    SQS_NAME_T, SQS_NAME,
    DLQ_NAME_T, DLQ_NAME,
    SQS_ARN_T
)

RES_KEY = f"x-{os.path.basename(os.path.dirname(os.path.abspath(__file__)))}"
SQS_SSM_PREFIX = f"/{RES_KEY}/"


def define_queue_tags(properties, queue_name):
    """
    Function to define the SQS Queue Tags

    :param properties: Properties imported from Compose File
    :type properties: dict
    :param queue_name: The Queue name as defined in ComposeX
    :type queue_name: str

    :returns: Tags()
    :rtype: troposphere.Tags
    """
    tag_name_exists = False
    if KEYISSET('Tags', properties):
        for key in properties['Tags']:
            if key == 'Name':
                tag_name_exists = True
    else:
        properties['Tags'] = {}

    if not tag_name_exists:
        properties['Tags']['Name'] = queue_name
    return Tags(properties['Tags'])


def add_redrive_policy(queue_tpl, queue_name, properties, dlq_name):
    """
    Function to add a DLQ redrive policy to the SQS Queue

    :param queue_tpl: Template of the queue to add the redrive for
    :type queue_tpl: troposphere.Template
    :param queue_name: Name of the Queue as defined in ComposeX file
    :type queue_name: str
    :param properties: Queue properties as defined in ComposeX file
    :type properties: dict
    :param dlq_name: Name of the redrive queue as defined in ComposeX File
    :type dlq_name: str

    :returns: dict containing the RedrivePolicy
    :rtype: dict
    """
    ssm_string = f"/${{{ROOT_STACK_NAME_T}}}{SQS_SSM_PREFIX}${{{DLQ_NAME_T}}}/{SQS_ARN_T}"
    ssm_resolve = Sub(r'{{resolve:ssm:%s:1}}' % (ssm_string))
    cfn_import = ImportValue(
        Sub(
            f"${{{ROOT_STACK_NAME_T}}}"
            f"-${{{DLQ_NAME_T}}}"
            f"-{SQS_ARN_T}"
        )
    )
    redrive_queue_import = If(
        USE_SSM_ONLY_T,
        ssm_resolve,
        cfn_import
    )
    queue_tpl.add_parameter(DLQ_NAME)
    policy = properties['RedrivePolicy']
    policy['deadLetterTargetArn'] = redrive_queue_import
    policy_obj = RedrivePolicy(**policy)
    return {'RedrivePolicy': policy_obj}


def add_ssm_parameters(src_template, queue_obj):
    """
    Function to add SSM parameters to store Queue information

    :param src_template: Template to add the SSM params to
    :type src_template: troposphere.Template
    :param queue_obj: Queue to ref/getatt from
    :type queue_obj: troposphere.sqs.Queue
    """
    ssm_name_string = f"/${{{ROOT_STACK_NAME_T}}}{SQS_SSM_PREFIX}${{{SQS_NAME_T}}}"
    LOG.debug(ssm_name_string)
    SsmParameter(
        'QueueArnSsmParameter',
        template=src_template,
        Condition=USE_SSM_EXPORTS_T,
        DependsOn=[queue_obj],
        Name=Sub(f"{ssm_name_string}/{SQS_ARN_T}"),
        Value=GetAtt(queue_obj, 'Arn'),
        Tier='Standard',
        Type='String'
    )
    SsmParameter(
        'QueueNameSsmParameter',
        template=src_template,
        Condition=USE_SSM_EXPORTS_T,
        DependsOn=[queue_obj],
        Name=Sub(f"{ssm_name_string}/{SQS_NAME_T}"),
        Value=GetAtt(queue_obj, 'QueueName'),
        Tier='Standard',
        Type='String'
    )


def generate_queue_template(queue_name, properties, redrive_queue=None, **kwargs):
    """
    Function that generates a single queue template

    :param queue_name: Name of the Queue as defined in ComposeX File
    :type queue_name: str
    :param properties: The queue properties
    :type properties: dict

    :returns: queue_template
    :rtype: troposphere.Template
    """
    if not queue_name:
        raise TypeError("Parameter queue_name must be a non-empty string")
    queue_template = build_template(
        f"Queue {queue_name} in {{{ROOT_STACK_NAME_T}}}",
        [SQS_NAME]
    )
    if redrive_queue is not None:
        properties.update(add_redrive_policy(
            queue_template, queue_name, properties, redrive_queue)
        )
    properties['Tags'] = define_queue_tags(properties, queue_name)
    if 'QueueName' in properties.keys():
        properties.pop('QueueName')
        properties['QueueName'] = Sub(
            f"${{{ROOT_STACK_NAME_T}}}-${properties['QueueName']}"
        )
    else:
        properties['QueueName'] = Sub(
            f"${{{ROOT_STACK_NAME_T}}}-${{{SQS_NAME_T}}}"
        )
    queue = Queue(queue_name, template=queue_template, **properties)
    add_ssm_parameters(queue_template, queue)
    cfn_prefix = f"${{{ROOT_STACK_NAME_T}}}-${{{SQS_NAME_T}}}"
    queue_template.add_output(formatted_outputs(
        [
            {SQS_NAME_T: GetAtt(queue, 'QueueName')},
            {SQS_ARN_T: GetAtt(queue, 'Arn')}
        ],
        export=True,
        prefix=cfn_prefix
    ))
    return queue_template


def add_queue_stack(queue_name, queue, queues, session, **kwargs):
    """
    Function to define the Queue template settings for the Nested Stack

    :param queue_name: Name of the queue as defined in Docker ComposeX file
    :param queue: the queue
    :param queues: all the queues in a list
    :param session: session to override
    :param kwargs: optional arguments

    :return: Queue Stack object
    :rtype: troposphere.cloudformation.Stack
    """
    depends_on = []
    parameters = {
        SQS_NAME_T: queue_name,
        ROOT_STACK_NAME_T: If(
            USE_STACK_NAME_CON_T,
            Ref('AWS::StackName'),
            Ref(ROOT_STACK_NAME)
        ),
        cfn_params.USE_CFN_EXPORTS_T: Ref(cfn_params.USE_CFN_EXPORTS),
        cfn_params.USE_SSM_EXPORTS_T: Ref(cfn_params.USE_SSM_EXPORTS)
    }
    if KEYPRESENT('Properties', queue):
        properties = queue['Properties']
    else:
        properties = {}
    if KEYISSET('RedrivePolicy', properties):
        redrive_target = properties['RedrivePolicy']['deadLetterTargetArn']
        if redrive_target not in queues:
            raise KeyError(
                f'Queue {redrive_target} defined as DLQ for {queue_name} but is not defined'
            )
        depends_on.append(redrive_target)
        parameters.update({DLQ_NAME_T: redrive_target})
        queue_tpl = generate_queue_template(
            queue_name,
            properties,
            redrive_target,
            **kwargs
        )
    else:
        queue_tpl = generate_queue_template(queue_name, properties, **kwargs)
    LOG.debug(parameters)
    LOG.debug(session)
    template_url = upload_template(
            template_body=queue_tpl.to_json(),
            bucket_name=kwargs['BucketName'],
            file_name=f'{queue_name}.json',
            session=session
    )
    queue_stack = Stack(
        queue_name,
        Parameters=parameters,
        DependsOn=depends_on,
        TemplateURL=template_url
    )
    return queue_stack


def generate_sqs_root_template(compose_content, session, **kwargs):
    """
    Generates a base template for a sqs queues. Iterates over each queue defined
    in x-sqs of the ComposeX file and identify settings and properties for these

    :param compose_content: The Docker compose content
    :type compose_content: dict
    :param session: boto3 session to override default
    :type session: boto3.session.Session

    :return: SQS Root/Parent template
    :rtype: troposphere.Template
    """
    validate_kwargs(
        ['BucketName'],
        kwargs
    )
    description = 'Root SQS Template'
    if KEYISSET('EnvName', kwargs):
        description = f"Root SQS Template for {kwargs['EnvName']}"
    root_tpl = build_template(description)
    queues = compose_content[RES_KEY]
    for queue_name in queues:
        LOG.debug(queue_name)
        LOG.debug(session)
        queue_stack = add_queue_stack(
            queue_name, queues[queue_name], queues.keys(), session, **kwargs
        )
        root_tpl.add_resource(queue_stack)
    return root_tpl



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Console script for ecs_composex."""

import sys
import argparse
import warnings
from boto3 import session
from ecs_composex.common import LOG
from ecs_composex.common.aws import (
    BUCKET_NAME, CURATED_AZS
)
from ecs_composex.common import KEYISSET
from ecs_composex.root import generate_full_template
from ecs_composex.vpc.vpc_params import (
    APP_SUBNETS_T, PUBLIC_SUBNETS_T, STORAGE_SUBNETS_T,
    VPC_ID_T, VPC_MAP_ID_T
)
from ecs_composex.ecs.ecs_params import LAUNCH_TYPE_T
from ecs_composex.compute.cluster_params import (
    CLUSTER_NAME_T, USE_FLEET_T
)


def validate_vpc_input(args):
    """Function to validate the VPC arguments are all present

    :param args: Parser arguments
    :type args: dict

    :raise: KeyError if missing argument when not creating VPC

    """
    nocreate_requirements = [
        PUBLIC_SUBNETS_T,
        APP_SUBNETS_T,
        STORAGE_SUBNETS_T,
        VPC_ID_T, VPC_MAP_ID_T
    ]
    if not KEYISSET('CreateVpc', args):
        for key in nocreate_requirements:
            if not KEYISSET(key, args):
                raise ValueError(f"If you want to use an existing VPC, you need {key}")
    else:
        for key in nocreate_requirements:
            if KEYISSET(key, args):
                LOG.info(args[key])
                warnings.warn(
                    f"Creating VPC is set but you also set {key}. Ignoring and using new VPC values",
                    UserWarning
                )


def validate_cluster_input(args):
    """Function to validate the cluster arguments

    :param args: Parser arguments
    """
    if not KEYISSET('CreateCluster', args) and not KEYISSET(CLUSTER_NAME_T, args):
        raise KeyError(f"You must provide an ECS Cluster name if you do not want ECS ComposeX to create one for you")


def main():
    """Console script for ecs_composex."""
    parser = argparse.ArgumentParser()
    #  Generic settings
    parser.add_argument(
        '-f', '--docker-compose-file', dest='ComposeXFile',
        required=True, help="Path to the Docker compose file"
    )
    parser.add_argument(
        '-o', '--output-file', type=str, required=True,
        help="The name and path of the main output file. If you specify extra arguments, it will create a parameters"
             " file as well for creating your CFN Stack"
    )
    #  AWS SETTINGS
    parser.add_argument(
        '--region', required=False, default=session.Session().region_name,
        dest='AwsRegion',
        help="Specify the region you want to build for"
             "default use default region from config or environment vars"
    )
    parser.add_argument(
        '--az', dest='AwsAzs', action='append', required=False, default=CURATED_AZS,
        help="List AZs you want to deploy to specifically within the region"
    )
    parser.add_argument(
        '-b', '--bucket-name', type=str, required=False, default=BUCKET_NAME,
        help='Bucket name to upload the templates to', dest='BucketName'
    )
    # VPC SETTINGS
    parser.add_argument(
        '--create-vpc', required=False, default=False, action='store_true',
        help="Create a VPC for this deployment", dest='CreateVpc'
    )
    parser.add_argument(
        '--vpc-cidr', required=False, default='192.168.36.0/22', dest='VpcCidr',
        help="Specify the VPC CIDR if you use --create-vpc"
    )
    parser.add_argument(
        '--vpc-id', dest=VPC_ID_T, required=False, type=str,
        help='Specify VPC ID when not creating one'
    )
    parser.add_argument(
        '--public-subnets', required=False, dest=PUBLIC_SUBNETS_T, action='append',
        help="List of Subnet IDs to use for the cluster when not creating VPC"
    )
    parser.add_argument(
        '--app-subnets', required=False, dest=APP_SUBNETS_T, action='append',
        help="List of Subnet IDs to use for the cluster when not creating VPC"
    )
    parser.add_argument(
        '--storage-subnets', required=False, dest=STORAGE_SUBNETS_T, action='append',
        help="List of Subnet IDs to use for the cluster when not creating VPC"
    )
    parser.add_argument(
        '--discovery-map-id', '--map', dest=VPC_MAP_ID_T, required=False, help="Service Discovery ID, ie. ns-xxx"
    )
    parser.add_argument(
        '--single-nat', dest='SingleNat', action='store_true',
        help="Whether you want a single NAT for your application subnets or not. Not recommended for production"
    )
    # CLUSTER SETTINGS
    parser.add_argument(
        '--create-cluster', required=False, default=False, action='store_true',
        help="Create an ECS Cluster for this deployment", dest='CreateCluster'
    )
    parser.add_argument(
        '--cluster-name', type=str, required=False, dest=CLUSTER_NAME_T,
        help='Override/Provide ECS Cluster name'
    )
    # COMPUTE SETTINGS
    parser.add_argument(
        '--use-fargate', required=False, default=False, action='store_true',
        dest=LAUNCH_TYPE_T, help="If you run Fargate only, no EC2 will be created"
    )
    parser.add_argument(
        '--use-spot-fleet', required=False, default=False, action='store_true',
        dest=USE_FLEET_T, help="Runs spotfleet for EC2. If used in combination "
                               "of --use-fargate, it will create an additional SpotFleet"
    )
    parser.add_argument(
        '--create-launch-template', dest='CreateLaunchTemplate', action='store_true',
        help='Whether you want to create a launch template to create EC2 resources for'
        ' to expand the ECS Cluster and run containers on EC2 instances you might have access to.'
    )
    #  ECS COMPOSEX SPECIALS
    parser.add_argument(
        '--iam-only', default=False, action='store_true', required=False,
        help="Generates only the IAM roles for the tasks"
    )

    parser.add_argument('_', nargs='*')
    args = parser.parse_args()

    validate_vpc_input(vars(args))
    validate_cluster_input(vars(args))

    print("Arguments: " + str(args._))
    generate_full_template(**vars(args))


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `ecs_composex` package."""

from troposphere import Template, Parameter

from ecs_composex.common import (
    init_template,
    build_template,
    add_parameters
)
from ecs_composex.vpc import vpc_params


def test_init_template():
    """
    Test that makes sure we get a new template
    """
    tpl = init_template()
    assert isinstance(tpl, Template)
    tpl = init_template('Override')
    assert tpl.description == 'Override'
    add_parameters(tpl, [Parameter('Test', Type='String')])
    assert len(tpl.parameters) == 1
    assert tpl.parameters['Test'].Type == 'String'
    add_parameters(tpl, [vpc_params.VPC_ID])
    assert len(tpl.parameters) == 2
    assert tpl.parameters[vpc_params.VPC_ID_T].Type == vpc_params.VPC_TYPE


def test_build_template():
    """
    Testing build template which is a merge of init and add_parameters
    """
    tpl = build_template(
        'Override',
        [
            vpc_params.VPC_ID,
            vpc_params.VPC_MAP_ID
        ]
    )
    assert isinstance(tpl, Template)
    assert isinstance(tpl.parameters[vpc_params.VPC_ID_T], Parameter)
    assert len(tpl.parameters) == 5
    assert tpl.description == 'Override'


# @pytest.fixture
# def response():
#     """Sample pytest fixture.

#     See more at: http://doc.pytest.org/en/latest/fixture.html
#     """
#     # import requests
#     # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


# def test_content(response):
#     """Sample pytest test function with the pytest fixture as an argument."""
#     # from bs4 import BeautifulSoup
#     # assert 'GitHub' in BeautifulSoup(response.content).title.string



# -*- coding: utf-8 -*-

"""
Functions to manage a template and wheter it should be stored in S3
"""

import boto3
from botocore.exceptions import ClientError

from ecs_composex.common import DATE_PREFIX
from ecs_composex.common import LOG


def check_bucket(bucket_name, session=None, client=None):
    """
    Function that checks if the S3 bucket exists and if not attempts to create it.

    :param bucket_name: name of the s3 bucket
    :type bucket_name: str
    :param session: boto3 session to use if wanted to override settings.
    :type session: boto3.session.Session
    :param client: override client
    :type client: boto3.client

    :returns: True/False, Returns whether the bucket exists or not for upload
    :rtype: bool
    """
    if session is None:
        session = boto3.session.Session()
    region = session.region_name
    location = {'LocationConstraint': region}
    if client is None:
        client = session.client('s3')
    try:
        client.head_bucket(
            Bucket=bucket_name
        )
        LOG.debug(f'Bucket {bucket_name} available')
        return True
    except ClientError as error:
        if error.response['Error']['Code'] == '404':
            try:
                LOG.info(f"Attempting to create bucket {bucket_name}")
                client.create_bucket(
                    ACL='private',
                    Bucket=bucket_name,
                    ObjectLockEnabledForBucket=True,
                    CreateBucketConfiguration=location
                )
                LOG.info(f"Bucket {bucket_name} successfully created.")
                return True
            except Exception as error:
                LOG.error(error)
                return False
    except Exception as error:
        LOG.error(error)
        LOG.error(f'Bucket name: {bucket_name}')
        LOG.error(type(bucket_name))
        return False


def validate_template(template_body, file_name, template_url=None, session=None):
    """
    Uses AWS CFN validate-template API to validate the template

    :param template_body: Template body, would come from troposphere template to_json() or to_yaml()
    :type template_body: str
    :param file_name: Name of the file
    :type file_name: str
    :param template_url: Template URL to check if template was uploaded to S3 already, optional
    :type template_url: str
    :param session: boto3 session to override and use for the client, optional
    :type session: boto3.session.Session() to override client

    :returns: True if template is validated by CFN, false if error
    :rtype: bool
    """
    if session is None:
        client = boto3.client('cloudformation')
    else:
        client = session.client('cloudformation')

    try:
        if template_url is None:
            client.validate_template(TemplateBody=template_body)
        else:
            client.validate_template(TemplateURL=template_url)
        return True
    except Exception as error:
        LOG.error(error)
        with open(f'/tmp/{file_name}', 'w') as fd:
            fd.write(template_body.strip('\n'))
        LOG.error(f'Non valid template successfully written to /tmp/{file_name}')
    return False


def upload_template(template_body, bucket_name, file_name, validate=True,
                    prefix=None, session=None, client=None, **kwargs):
    """Upload template_body to a file in s3 with given prefix and bucket_name

    :param template_body: Template body, would come from troposphere template to_json() or to_yaml()
    :type template_body: str
    :param bucket_name: name of the bucket to upload the file to
    :type bucket_name: str
    :param file_name: Name of the file
    :type file_name: str
    :param prefix: override default prefix for the file in S3
    :type prefix: str, optional
    :param validate: Whether you want to validate the uploaded template to S3. Default: True
    :type validate: bool, optional
    :param session: boto3 session to override and use for the client, optional
    :type session: boto3.session.Session() to override client
    :param client: client override for call
    :type client: boto3.client

    :returns: url_path, the https://s3.amazonaws.com/ URL to the file
    :rtype: str
    """
    if session is None:
        session = boto3.session.Session()
    assert check_bucket(bucket_name=bucket_name, session=session)

    if prefix is None:
        prefix = DATE_PREFIX

    key = f'{prefix}/{file_name}'
    if client is None:
        client = session.client('s3')
    try:
        client.put_object(
            Body=template_body,
            Key=key,
            Bucket=bucket_name,
            ContentEncoding='utf-8',
            ContentType='application/json',
            ServerSideEncryption='AES256',
            **kwargs
        )
        url_path = f'https://s3.amazonaws.com/{bucket_name}/{key}'
        if validate:
            assert validate_template(
                template_body, file_name=file_name,
                template_url=url_path, session=session
            )
        return url_path
    except Exception as error:
        LOG.debug(error)
        return None



# -*- coding: utf-8 -*-
""" Core ECS Template building """

from troposphere import Ref, Sub, Tags
from troposphere.cloudformation import Stack
from troposphere.ec2 import SecurityGroup
from troposphere.logs import LogGroup

from ecs_composex.common import LOG
from ecs_composex.common import (
    cfn_params, build_template, KEYISSET
)
from ecs_composex.common.cfn_params import (
    ROOT_STACK_NAME_T,
    ROOT_STACK_NAME
)
from ecs_composex.ecs import ecs_params
from ecs_composex.ecs.ecs_params import CLUSTER_NAME, CLUSTER_NAME_T
from ecs_composex.ecs.ecs_service import (
    generate_service_template
)
from ecs_composex.vpc import vpc_params


def validate_labels(service_labels):
    """
    Function to validate service labels
    # TO IMPLEMENT
    """
    return True


def validate_input(services):
    """
    Validates services docker format
    """
    props_must_have = [
        'image'
    ]
    for service_name in services:
        service = services[service_name]
        for prop in props_must_have:
            if not KEYISSET(prop, service):
                raise KeyError('Service {service_name} is missing property {prop}')
        if prop == 'labels':
            assert validate_labels(service[prop])
    return True


def add_clusterwide_security_group(template):
    """
    Function to generate the service Load Balancers (if Any)
    """
    sg = SecurityGroup(
        'ClusterWideSecurityGroup',
        template=template,
        GroupDescription=Sub(f"SG for ${{{CLUSTER_NAME_T}}}"),
        GroupName=Sub(f"cluster-${{{CLUSTER_NAME_T}}}"),
        Tags=Tags(
            {
                'Name': Sub(f"clustersg-${{{CLUSTER_NAME_T}}}"),
                'ClusterName': Ref(CLUSTER_NAME)
            }
        ),
        VpcId=Ref(vpc_params.VPC_ID)
    )
    return sg


def add_services_stacks(compose_content, root_tpl, cluster_sg, session=None, **kwargs):
    """Function putting together the ECS Service template

    :param compose_content: Docker/ComposeX file content
    :type compose_content: dict
    :param root_tpl: template
    :type root_tpl: troposphere.Template
    :param cluster_sg: cluster default security group
    :type cluster_sg: troposphere.ec2.SecurityGroup
    :param session: override default session
    :type session: boto3.session.Session
    :param kwargs: optional arguments
    :type kwargs: dicts or set
    """
    for service_name in compose_content[ecs_params.RES_KEY]:
        service = compose_content[ecs_params.RES_KEY][service_name]
        service_set = generate_service_template(
            compose_content,
            service_name,
            service,
            session=session,
            **kwargs
        )
        parameters = {
            CLUSTER_NAME_T: Ref(CLUSTER_NAME),
            ROOT_STACK_NAME_T: Ref(ROOT_STACK_NAME),
            ecs_params.CLUSTER_SG_ID_T: Ref(cluster_sg)
        }
        dependencies = [ecs_params.LOG_GROUP_T]
        LOG.debug(service_set[-1])
        if isinstance(service_set[-1], list):
            dependencies = dependencies + service_set[-1]
        LOG.debug(dependencies)
        parameters.update(service_set[1])
        if service_set[0]:
            Stack(
                service_name,
                template=root_tpl,
                TemplateURL=service_set[0],
                Parameters=parameters,
                DependsOn=dependencies
            )
        else:
            Warning(
                f"Template for service {service_name}"
                "was not successfully generated"
            )
        LOG.debug(f"Service {service_name} added.")
        LOG.info(f"Template URL: {service_set[0]}")


def generate_services_templates(compose_content, session=None, **kwargs):
    """
    Function to generate the root template
    """
    parameters = [
        cfn_params.SERVICE_DISCOVERY,
        CLUSTER_NAME,
        vpc_params.VPC_ID,
        vpc_params.PUBLIC_SUBNETS,
        vpc_params.APP_SUBNETS
    ]
    template = build_template(
        'Root template for ECS Services',
        parameters
    )

    template.add_resource(LogGroup(
        ecs_params.LOG_GROUP_T,
        RetentionInDays=30,
        LogGroupName=Ref(CLUSTER_NAME)
    ))
    cluster_sg = add_clusterwide_security_group(template)
    add_services_stacks(compose_content, template, cluster_sg, session=session, **kwargs)
    if KEYISSET('Debug', kwargs):
        with open('/tmp/ecs_root.json', 'w') as services_fd:
            services_fd.write(template.to_json())
    return template



# -*- coding: utf-8 -*-
"""
Core module for ECS ComposeX.

This module is going to parse each service and each x-resource key from the compose file
(hence ComposeX) and determine its

* ServiceDefinition
* TaskDefinition
* TaskRole
* ExecutionRole

It is going to also, based on the labels set in the compose file

* Add the service to Service Discovery via AWS CloudMap
* Add load-balancers to dispatch traffic to the microservice

"""

import boto3

from ecs_composex.common import load_composex_file
from ecs_composex.ecs.ecs_template import generate_services_templates


def create_services_templates(session=None, **kwargs):
    """
    :return:
    """
    if session is None:
        session = boto3.session.Session()
    content = load_composex_file(kwargs['ComposeXFile'])
    services_template = generate_services_templates(
        compose_content=content, session=session,
        **kwargs
    )
    return services_template



"""Entrypoint for IAM"""

from troposphere import Sub


def service_role_trust_policy(service_name):
    """
    Simple function to format the trust relationship for a Role and an AWS Service
    used from lambda-my-aws/ozone

    :param service_name: name of the service
    :type service_name: str

    :return: policy document
    :rtype: dict
    """
    statement = {
        "Effect": "Allow",
        "Principal": {
            "Service": [
                Sub(f'{service_name}.${{AWS::URLSuffix}}')
            ]
        },
        "Action": ["sts:AssumeRole"],
        "Condition": {
            "Bool": {
                "aws:SecureTransport": "true"
            }
        }
    }
    policy_doc = {
        "Version": "2012-10-17",
        "Statement": [
            statement
        ]
    }
    return policy_doc



# -*- coding: utf-8 -*-
""""
Common parameters for CFN
This is a crucial part as all the titles, maked `_T` are string which are then used the same way
across all imports, which gives consistency for CFN to use the same names,
which it heavily relies onto.

You can change the names *values* so you like so long as you keep it Alphanumerical [a-zA-Z0-9]
"""

from troposphere import Parameter


ROOT_STACK_NAME_T = 'RootStackName'
ROOT_STACK_NAME = Parameter(
    ROOT_STACK_NAME_T, Type='String', Default='<self>',
    Description='When part of a combined deployment, represents to the top stack name'
)

VPC_MAP_ID_T = 'AwsVpcCloudMapId'
VPC_MAP_ID = Parameter(VPC_MAP_ID_T, Type='String', Default='none')


VPC_MAP_ARN_T = 'AwsVpcCloudMapArn'
VPC_MAP_ARN = Parameter(VPC_MAP_ARN_T, Type='String', Default='none')

SERVICE_DISCOVERY_T = 'UseAwsCloudMap'
SERVICE_DISCOVERY = Parameter(
    SERVICE_DISCOVERY_T,
    Type='String',
    AllowedValues=['True', 'False'],
    Default='True'
)

USE_CFN_PARAMS_T = 'UseCfnParametersValue'
USE_CFN_PARAMS = Parameter(
    USE_CFN_PARAMS_T,
    Type='String',
    AllowedValues=['True', 'False'], Default=True
)

USE_CFN_EXPORTS_T = 'UseCfnExports'
USE_CFN_EXPORTS = Parameter(
    USE_CFN_EXPORTS_T, Type='String',
    AllowedValues=['True', 'False'], Default='True'
)

USE_SSM_EXPORTS_T = 'UseSsmExports'
USE_SSM_EXPORTS = Parameter(
    USE_SSM_EXPORTS_T, Type='String',
    AllowedValues=['True', 'False'], Default='False'
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `ecs_composex` package."""

import os
import yaml
import pytest
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from ecs_composex.ecs_composex import generate_x_resource_configs

try:
    BUCKET = os.environ['KNOWN_BUCKET']
except KeyError:
    BUCKET = 'ecs-composex-dev'

HERE = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def content():
    """
    Opens the file and passes the content around
    """
    with open(f"{HERE}/services_with_queues.yml", 'r') as fd:
        content = yaml.load(fd.read(), Loader=Loader)
    return content


@pytest.fixture
def args():
    return {
        'BucketName': BUCKET,
        'EnvName': 'abcd',
        'Debug': True
    }


def test_x_resource_permissions(content, args):
    """
    Test generating the permissions for all x- resources
    """
    permissions = generate_x_resource_configs(content, **args)
    assert 'x-sqs' in permissions.keys()



# -*- coding: utf-8 -*-

"""
Module to create the Launch Template for the hosts and the associated security group and
IAM Role (with Instance Profile).

The Launch Template uses CFN-INIT in order to bootstrap the machine at runtime, avoiding
any hardcoding, especially not for the ECS Service

These settings are all documented on AWS official documentation:
https://docs.aws.amazon.com/AmazonECS/latest/developerguide/ecs-agent-config.html
"""

from troposphere import (
    cloudformation, Tags, Sub,
    Base64, Ref, GetAtt, Join, If
)
from troposphere.ec2 import (
    Monitoring,
    IamInstanceProfile
)
from troposphere.ec2 import (
    SecurityGroup,
    LaunchTemplate,
    LaunchTemplateData,
    TagSpecifications,
    EBSBlockDevice,
    LaunchTemplateBlockDeviceMapping
)
from troposphere.iam import (
    Role,
    Policy,
    InstanceProfile
)

from ecs_composex.compute import cluster_params
from ecs_composex.compute.cluster_params import (
    HOST_PROFILE_T,
    HOST_ROLE_T,
    NODES_SG_T
)
from ecs_composex.common.cfn_conditions import USE_STACK_NAME_CON_T
from ecs_composex.common.cfn_params import ROOT_STACK_NAME_T, ROOT_STACK_NAME
from ecs_composex.iam import service_role_trust_policy
from ecs_composex.vpc import vpc_params
from ecs_composex.ecs.ecs_params import CLUSTER_NAME_T


def add_hosts_profile(template):
    """
    Adds role to the template

    :parm template: EC2 Cluster template to add the role and profile to
    :type template: troposphere.Template

    :returns: troposphere IAM Role for EC2 hosts
    :rtype: troposphere.iam.Role
    """
    ecs_policy = Policy(
        PolicyName='AllowEcsSpecific',
        PolicyDocument={
            'Version': '2012-10-17',
            'Statement': [
                {
                    'Effect': 'Allow',
                    'Action': [
                        "ecs:RegisterContainerInstance",
                        "ecs:UpdateContainerInstancesState",
                        "ecs:DeregisterContainerInstance"
                    ],
                    'Resource': [
                        Sub(
                            "arn:${AWS::Partition}:ecs:${AWS::Region}:${AWS::AccountId}:"
                            f"cluster/${{{ROOT_STACK_NAME_T}}}"
                        )
                    ]
                },
                {
                    'Effect': 'Allow',
                    'Action': [
                        "ecs:StartTelemetrySession",
                        "ecs:DiscoverPollEndpoint",
                        "ecs:Submit*",
                        "ecs:Poll"
                    ],
                    'Resource': ['*']
                }
            ]
        }
    )
    role = Role(
        HOST_ROLE_T,
        template=template,
        AssumeRolePolicyDocument=service_role_trust_policy('ec2'),
        ManagedPolicyArns=[
            "arn:aws:iam::aws:policy/service-role/AmazonEC2RoleforSSM"
        ],
        Policies=[ecs_policy]
    )
    InstanceProfile(
        HOST_PROFILE_T,
        template=template,
        Roles=[
            Ref(role)
        ],
    )
    return role


def add_hosts_security_group(template):
    """
    Function to add a security group for the host

    :parm template: EC2 Cluster template to add the SG to
    :type template: troposphere.Template

    :returns: troposphere IAM Role for EC2 hosts
    :rtype: troposphere.ec2.SecurityGroup
    """
    return SecurityGroup(
        NODES_SG_T,
        template=template,
        GroupDescription=Sub(f"Group for hosts in ${{{ROOT_STACK_NAME_T}}}"),
        VpcId=Ref(vpc_params.VPC_ID)
    )


def add_launch_template(template, hosts_sg):
    """Function to create a launch template.

    :param template: ECS Cluster template
    :type template: troposphere.Template
    :param hosts_sg: security group for the EC2 hosts
    :type hosts_sg: troposphere.ec2.SecurityGroup

    :return: launch_template
    :rtype: troposphere.ec2.LaunchTemplate
    """
    # from troposphere.cloudformation import (
    #     WaitCondition, WaitConditionHandle
    # )
    #  Deactivated conditions given you could run with no EC2 at all.
    # Tricky condition to do as the WaitCondition and Handler cannot be created on a CFN Update, but only at the
    # very creation of the stack.
    # wait_handle = WaitConditionHandle(
    #     'BootstrapHandle',
    #     template=template
    # )
    # WaitCondition(
    #     'BootStrapCondition',
    #     template=template,
    #     DependsOn=[hosts_role],
    #     Handle=Ref(wait_handle),
    #     Timeout='900'
    # )

    launch_template = LaunchTemplate(
        'LaunchTemplate',
        template=template,
        Metadata=cloudformation.Metadata(
            cloudformation.Init(
                cloudformation.InitConfigSets(
                    default=[
                        'vmpackages',
                        'dockerconfig',
                        'ecsconfig',
                        'vmservices'
                    ]
                ),
                vmpackages=cloudformation.InitConfig(
                    packages={
                        'yum': {
                            "awslogs": [],
                            "amazon-ssm-agent": []
                        }
                    },
                    commands={
                        '001-check-packages': {
                            'command': 'rpm -qa | grep amazon'
                        },
                        '002-check-packages': {
                            'command': 'rpm -qa | grep aws'
                        }
                    }
                ),
                vmservices=cloudformation.InitConfig(
                    services={
                        "sysvinit": {
                            "amazon-ssm-agent": {
                                "enabled": True,
                                "ensureRunning": True
                            }
                        }
                    }
                ),
                dockerconfig=cloudformation.InitConfig(
                    commands={
                        '001-stop-docker': {
                            'command': 'systemctl stop docker'
                        },
                        '098-reload-systemd': {
                            'command': 'systemctl daemon-reload'
                        }
                    },
                    files={
                        '/etc/sysconfig/docker': {
                            'owner': 'root',
                            'group': 'root',
                            'mode': '644',
                            'content': Join(
                                '\n',
                                [
                                    'DAEMON_MAXFILES=1048576',
                                    Join(
                                        ' ',
                                        [
                                            "OPTIONS=--default-ulimit nofile=1024:4096"
                                        ]
                                    ),
                                    'DAEMON_PIDFILE_TIMEOUT=10',
                                    '#EOF',
                                    ''
                                ]
                            )
                        }
                    },
                    services={
                        'sysvinit': {
                            'docker': {
                                'enabled': True,
                                'ensureRunning': True,
                                'files': [
                                    '/etc/sysconfig/docker'
                                ],
                                'commands': [
                                    '098-reload-systemd'
                                ]
                            }
                        }
                    }
                ),
                ecsconfig=cloudformation.InitConfig(
                    files={
                        '/etc/ecs/ecs.config': {
                            'owner': 'root',
                            'group': 'root',
                            'mode': '644',
                            'content': Join('\n', [
                                Sub(f"ECS_CLUSTER=${{{CLUSTER_NAME_T}}}"),
                                'ECS_ENABLE_TASK_IAM_ROLE=true',
                                'ECS_ENABLE_SPOT_INSTANCE_DRAINING=true',
                                'ECS_ENABLE_TASK_IAM_ROLE_NETWORK_HOST=true',
                                'ECS_ENABLE_CONTAINER_METADATA=true',
                                'ECS_ENABLE_UNTRACKED_IMAGE_CLEANUP=true',
                                'ECS_UPDATES_ENABLED=true',
                                'ECS_ENGINE_TASK_CLEANUP_WAIT_DURATION=15m',
                                'ECS_IMAGE_CLEANUP_INTERVAL=10m',
                                'ECS_NUM_IMAGES_DELETE_PER_CYCLE=100',
                                'ECS_ENABLE_TASK_ENI=true',
                                'ECS_AWSVPC_BLOCK_IMDS=true',
                                'ECS_TASK_METADATA_RPS_LIMIT=300,400',
                                'ECS_ENABLE_AWSLOGS_EXECUTIONROLE_OVERRIDE=true',
                                'ECS_AVAILABLE_LOGGING_DRIVERS=["awslogs", "json-file"]',
                                '#EOF'
                            ])
                        }
                    },
                    commands={
                        '0001-restartecs': {
                            'command': 'systemctl --no-block restart ecs'
                        }
                    }
                )
            )
        ),
        LaunchTemplateData=LaunchTemplateData(
            BlockDeviceMappings=[
                LaunchTemplateBlockDeviceMapping(
                    DeviceName='/dev/xvda',
                    Ebs=EBSBlockDevice(
                        DeleteOnTermination=True,
                        Encrypted=True
                    )
                )
            ],
            ImageId=Ref(cluster_params.ECS_AMI_ID),
            InstanceInitiatedShutdownBehavior='terminate',
            IamInstanceProfile=IamInstanceProfile(
                Arn=Sub(f"${{{HOST_PROFILE_T}.Arn}}")
            ),
            TagSpecifications=[
                TagSpecifications(
                    ResourceType='instance',
                    Tags=Tags(
                        Name=Sub(f"EcsNodes-${{{ROOT_STACK_NAME_T}}}"),
                        StackName=Ref('AWS::StackName'),
                        StackId=Ref('AWS::StackId')
                    )
                )
            ],
            InstanceType='m5a.large',
            Monitoring=Monitoring(Enabled=True),
            SecurityGroupIds=[
                GetAtt(hosts_sg, 'GroupId')
            ],
            UserData=Base64(Join('\n', [
                '#!/usr/bin/env bash',
                'export PATH=$PATH:/opt/aws/bin',
                'cfn-init -v || yum install aws-cfn-bootstrap -y',
                Sub(f'cfn-init --region ${{AWS::Region}} -r LaunchTemplate -s ${{AWS::StackName}}'),
                # 'if [ $? -ne 0 ]; then',
                # Sub(f'cfn-signal -e 1 -r "Failed to bootstrap" \'${{{wait_handle.title}}}\''),
                # 'halt',
                # 'else',
                # Sub(f'cfn-signal -e 0 -r "Successfully bootstrapped" \'${{{wait_handle.title}}}\''),
                # 'fi',
                '# EOF'
            ]))
        ),
        LaunchTemplateName=If(
            USE_STACK_NAME_CON_T,
            Ref('AWS::StackName'),
            Ref(ROOT_STACK_NAME)
        )
    )
    return launch_template


def add_hosts_resources(template):
    """Function to add the LaunchTemplate, SG and IAM Profile to go along with the ECS Cluster

    :param template: the ecs_cluster template to add the hosts config to
    :type template: troposphere.Template
    :param cluster: the cluster to add the hosts to
    :type cluster: troposphere.ecs.Cluster

    :returns: launch_template
    """
    hosts_sg = add_hosts_security_group(template)
    add_hosts_profile(template)
    launch_template = add_launch_template(template, hosts_sg)
    return launch_template



# -*- coding: utf-8 -*-
"""
Module to generate specific rules and otherwise configurations to allow communication between the
microservices
"""
import re

from troposphere import (
    Ref, Sub, GetAtt
)
from troposphere.ec2 import (
    SecurityGroupIngress
)

from ecs_composex.common import KEYISSET, LOG
from ecs_composex.ecs.ecs_params import (
    SERVICE_NAME_T, get_import_service_group_id,
    SG_T, RES_KEY
)

CIDR_REG = r"""((((((([0-9]{1}\.))|([0-9]{2}\.)|
(1[0-9]{2}\.)|(2[0-5]{2}\.)))){3})(((((([0-9]{1}))|
([0-9]{2})|(1[0-9]{2})|(2[0-5]{2}))))){1,3})\/(([0-9])|([1-2][0-9])|((3[0-2])))$"""
CIDR_PAT = re.compile(CIDR_REG)


def flatten_ip(ip_str):
    """Function to remove all non alphanum characters from IP CIDR notation

    :param ip_str:

    :rtype: str
    """
    return ip_str.replace('.', '').split('/')[0].strip()


def add_lb_to_service_ingress(template, lb_sg, service_sg, settings):
    """Function to add Service ingress between the LB and the microservice

    :param template: microservice template
    :param lb_sg:
    :type lb_sg: troposphere.ec2.SecurityGroup
    :param service_sg: security group of the microservice
    :type service_sg: str or troposphere.ec2.SecurityGroup
    :param settings: network settings as defined in compile_network_settings
    :type settings: dict
    """
    LOG.debug(f"Adding ALB ingress to service")
    for port in settings['ports']:
        SecurityGroupIngress(
            f"From{settings['lb_type'].title()}ToServicePort{port['target']}",
            template=template,
            FromPort=port['target'],
            ToPort=port['target'],
            IpProtocol=port['protocol'],
            GroupId=GetAtt(service_sg, 'GroupId'),
            SourceSecurityGroupOwnerId=Ref('AWS::AccountId'),
            SourceSecurityGroupId=GetAtt(lb_sg, 'GroupId'),
            Description=Sub(f"From LB to ${{{SERVICE_NAME_T}}} on port {port['target']}")
        )


def add_public_security_group_ingress(template, service_name, settings, security_group):
    """Function to add public ingress. If a list of IPs is found in the config['ext_sources']
    then it will use that, if not, allows from 0.0.0.0/0

    :param template: service template to add the ingress rules to
    :type template: troposphere.Template
    :param service_name: name of the service
    :type service_name: str
    :param settings: network settings as defined in compile_network_settings
    :type settings: dict
    :param security_group: security group (object or title string) to add the rules to
    :type security_group: str or troposphere.ec2.SecurityGroup
    """
    if KEYISSET('ext_sources', settings) and isinstance(settings['ext_sources'], list):
        allowed_sources = settings['ext_sources']
    else:
        allowed_sources = [{'ipv4': '0.0.0.0/0', 'protocol': -1, 'source_name': 'ANY'}]

    for allowed_source in allowed_sources:
        props = {}
        if not KEYISSET('ipv4', allowed_source) and not KEYISSET('ipv6', allowed_source):
            LOG.warn('No IPv4 or IPv6 set. Skipping')
            continue

        props['CidrIp'] = allowed_source['ipv4'] if KEYISSET('ipv4', allowed_source) else Ref('AWS::NoValue')
        props['CidrIpv6'] = allowed_source['ipv6'] if KEYISSET('ipv6', allowed_source) else Ref('AWS::NoValue')

        if KEYISSET('CidrIp', props) and isinstance(props['CidrIp'], str) and not CIDR_PAT.match(props['CidrIp']):
            LOG.error(f"Falty IP Address: {allowed_source} - service {service_name}")
            raise ValueError('Not a valid IPv4 CIDR notation', props['CidrIp'], 'Expected', CIDR_REG)

        LOG.debug(f"Adding {allowed_source} for ingress")

        for port in settings['ports']:
            if KEYISSET('source_name', allowed_source):
                title = f"From{allowed_source['source_name'].title()}Onto{port['target']}{port['protocol']}"
                description = Sub(
                    f"From {allowed_source['source_name'].title()} "
                    f"To {port['target']}{port['protocol']} for ${{{SERVICE_NAME_T}}}"
                )
            else:
                title = f"From{flatten_ip(allowed_source['ipv4'])}" \
                        "To{port['target']}{port['protocol']}"
                description = Sub(
                    f"Public {port['target']}{port['protocol']}"
                    f" for ${{{SERVICE_NAME_T}}}"
                )
            SecurityGroupIngress(
                title,
                template=template,
                Description=description,
                GroupId=GetAtt(security_group, 'GroupId'),
                IpProtocol=port['protocol'],
                FromPort=port['target'],
                ToPort=port['target'],
                **props
            )


def define_service_to_service_ingress(compose_content, template, service_name, service):
    """
    Function to determine the security group openings from a service to another

    :param compose_content: docker compose dictionary
    :type compose_content: dict
    :param template: the service template to add the resources to
    :type template: troposphere.Template
    :param service_name: name of the service as defined in Docker ComposeX file
    :type service_name str
    :param service: the service definition
    :type service: dict

    :return: depends_on, list of dependencies for the *Stack* object
    :rtype: list
    """
    from ecs_composex.ecs.ecs_networking import compile_network_settings
    depends_on = []
    links = service['links'] if KEYISSET('links', service) else []
    for dest_service in links:
        if KEYISSET(dest_service, compose_content[RES_KEY]):
            target_service = compose_content[RES_KEY][dest_service]
            if KEYISSET('ports', target_service):
                network_settings = compile_network_settings(
                    compose_content, target_service, dest_service
                )
                LOG.debug(network_settings)
                LOG.debug(f'Creating ingress rules for {service_name} to access {dest_service}')
                depends_on.append(dest_service)
                for port in network_settings['ports']:
                    SecurityGroupIngress(
                        f"From{service_name}To{dest_service}Port{port['target']}",
                        template=template,
                        SourceSecurityGroupOwnerId=Ref('AWS::AccountId'),
                        SourceSecurityGroupId=GetAtt(SG_T, 'GroupId'),
                        IpProtocol=port['protocol'],
                        FromPort=port['target'],
                        ToPort=port['target'],
                        GroupId=get_import_service_group_id(dest_service)
                    )
    LOG.debug(depends_on)
    return depends_on



# -*- coding: utf-8 -*-
""" Parameters for SQS"""

from troposphere import Parameter

SQS_ARN_T = 'SqsQueueArn'
SQS_ARN = Parameter(SQS_ARN_T, Type='String')


SQS_NAME_T = 'QueueName'
SQS_NAME = Parameter(SQS_NAME_T, Type='String')

DLQ_NAME_T = 'DeadLetterQueueName'
DLQ_NAME = Parameter(DLQ_NAME_T, Type='String')



# -*- coding: utf-8 -*-

"""
Create the VPC template and its associated resources
TODO : Implement VPC Endpoints, NetworkACLs, VPC Flow logging to S3.
"""

import boto3
from troposphere import (
    GetAtt, Tags,
    Join, Ref, Sub, If
)
from troposphere.ec2 import (
    VPC as VPCType, VPCGatewayAttachment,
    InternetGateway,
    DHCPOptions, VPCDHCPOptionsAssociation
)
from troposphere.route53 import HostedZoneVPCs, HostedZone
from troposphere.servicediscovery import PrivateDnsNamespace as VpcSpace

from ecs_composex.common import (
    cfn_params,
    cfn_conditions,
    build_template
)
from ecs_composex.common.cfn_conditions import (
    USE_CLOUDMAP_CON_T,
    USE_STACK_NAME_CON_T
)
from ecs_composex.common.cfn_params import (
    ROOT_STACK_NAME,
    ROOT_STACK_NAME_T
)
from ecs_composex.common.outputs import formatted_outputs
from ecs_composex.common.templates import validate_template
from ecs_composex.vpc import (
    vpc_params, aws_mappings, vpc_conditions
)
from ecs_composex.vpc.vpc_maths import get_subnet_layers
from ecs_composex.vpc.vpc_subnets import (
    add_public_subnets,
    add_storage_subnets,
    add_apps_subnets
)

VPC_T = 'Vpc'
IGW_T = 'InternetGatewayV4'


def add_cloudmap_support(template, vpc):
    """
    Function add a VPC CloudMap instance

    :param template: VPC Template()
    :param vpc: Vpc() object for Ref()

    :returns: NIL
    """
    map_id = VpcSpace(
        'VpcCloudMapNameSpace',
        Condition=USE_CLOUDMAP_CON_T,
        template=template,
        Description=Sub(f"Map for VPC ${{{vpc.title}}}"),
        Vpc=Ref(vpc),
        Name=If(
            USE_STACK_NAME_CON_T,
            Sub(f'${{AWS::StackName}}.${{{vpc_params.VPC_DNS_ZONE_T}}}'),
            Sub(f'${{{ROOT_STACK_NAME_T}}}.${{{vpc_params.VPC_DNS_ZONE_T}}}')
        )
    )
    template.add_output(
        formatted_outputs(
            [
                {vpc_params.VPC_MAP_ID_T: GetAtt(map_id, 'Id')},
                {vpc_params.VPC_MAP_ARN_T: GetAtt(map_id, 'Arn')}
            ],
            export=True
        )
    )


def add_template_outputs(template, vpc, storage_subnets, public_subnets, app_subnets):
    """
    Function to add outputs / exports to the template

    :param template: VPC Template()
    :param vpc: Vpc() for Ref()
    :param storage_subnets: List of Subnet()
    :param public_subnets: List of Subnet()
    :param app_subnets: List of Subnet()

    :returns: NIL
    """
    template.add_output(formatted_outputs(
        [
            {vpc_params.VPC_ID_T: Ref(vpc)},
            {
                vpc_params.STORAGE_SUBNETS_T: Join(
                    ',', [Ref(subnet) for subnet in storage_subnets]
                )
            },
            {
                vpc_params.PUBLIC_SUBNETS_T: Join(
                    ',', [Ref(subnet) for subnet in public_subnets]
                )
            },
            {
                vpc_params.APP_SUBNETS_T: Join(
                    ',', [Ref(subnet) for subnet in app_subnets]
                )
            }
        ],
        export=True
    ))


def add_vpc_cidrs_outputs(template, layers):
    """
    Function to add outputs / exports to the template

    :param template: VPC Template()
    :type template: troposphere.Template
    :param layers: dict of layers CIDRs to export
    :type layers: dict
    """
    template.add_output(formatted_outputs(
        [
            {
                vpc_params.STORAGE_SUBNETS_CIDR_T: Join(
                    ',', [cidr for cidr in layers['stor']]
                )
            },
            {
                vpc_params.PUBLIC_SUBNETS_CIDR_T: Join(
                    ',', [cidr for cidr in layers['pub']]
                )
            },
            {
                vpc_params.APP_SUBNETS_CIDR_T: Join(
                    ',', [cidr for cidr in layers['app']]
                )
            }
        ],
        export=True
    ))


def add_vpc_core(template, vpc_cidr):
    """
    Function to create the core resources of the VPC
    and add them to the core VPC template

    :param template: VPC Template()
    :param vpc_cidr: str of the VPC CIDR i.e. 192.168.0.0/24

    :return: tuple() with the vpc and igw object
    """
    vpc = VPCType(
        VPC_T,
        template=template,
        CidrBlock=vpc_cidr,
        EnableDnsHostnames=True,
        EnableDnsSupport=True,
        Tags=Tags(
            Name=If(
                USE_STACK_NAME_CON_T,
                Ref('AWS::StackName'),
                Ref(ROOT_STACK_NAME)
            ),
            EnvironmentName=If(
                USE_STACK_NAME_CON_T,
                Ref('AWS::StackName'),
                Ref(ROOT_STACK_NAME)
            )
        )
    )
    igw = InternetGateway(
        IGW_T,
        template=template
    )
    VPCGatewayAttachment(
        "VPCGatewayAttachement",
        template=template,
        InternetGatewayId=Ref(igw),
        VpcId=Ref(vpc)
    )
    dhcp_opts = DHCPOptions(
        'VpcDhcpOptions',
        template=template,
        DomainName=If(
            USE_STACK_NAME_CON_T,
            Sub(f'${{AWS::StackName}}.${{{vpc_params.VPC_DNS_ZONE_T}}}'),
            Sub(f'${{{ROOT_STACK_NAME_T}}}.${{{vpc_params.VPC_DNS_ZONE_T}}}')
        ),
        DomainNameServers=['AmazonProvidedDNS'],
        Tags=Tags(
            Name=Sub(f'dhcp-${{{vpc.title}}}')
        )
    )
    VPCDHCPOptionsAssociation(
        'VpcDhcpOptionsAssociate',
        template=template,
        DhcpOptionsId=Ref(dhcp_opts),
        VpcId=Ref(vpc)
    )
    zone_con = template.add_condition(
        vpc_conditions.USE_SUB_ZONE_CON_T, vpc_conditions.USE_SUB_ZONE_CON
    )
    HostedZone(
        'VpcHostedZone',
        template=template,
        Condition=zone_con,
        VPCs=[
            HostedZoneVPCs(
                VPCId=Ref(vpc),
                VPCRegion=Ref('AWS::Region')
            )
        ],
        Name=If(
            USE_STACK_NAME_CON_T,
            Sub(f'sub.${{AWS::StackName}}.${{{vpc_params.VPC_DNS_ZONE_T}}}'),
            Sub(f'sub.${{{ROOT_STACK_NAME_T}}}.${{{vpc_params.VPC_DNS_ZONE_T}}}')
        ),
        HostedZoneTags=Tags(Name=Sub(f'ZoneFor-${{{vpc.title}}}'))
    )
    return (vpc, igw)


def generate_vpc_template(cidr_block, azs, session=None, single_nat=False):
    """
    Function to generate a new VPC template for CFN

    :param cidr_block: str of the CIDR used for this VPC
    :param azs: list of AWS Azs i.e. ['eu-west-1a', 'eu-west-1b']
    :param session: override session from boto3.session.Sesssion().
    :param single_nat: True/False if you want a single NAT for the Application Subnets
    :type single_nat: bool

    :return: Template() representing the VPC and associated resources
    """
    if session is None:
        session = boto3.session.Session()
    azs_count = len(azs)
    az_range = range(0, azs_count)
    layers = get_subnet_layers(cidr_block, len(azs))
    template = build_template(
        'VpcTemplate generated via ECS Compose X',
        [
            cfn_params.SERVICE_DISCOVERY,
            vpc_params.VPC_DNS_ZONE,
            vpc_params.USE_SUB_ZONE
        ]
    )
    template.add_mapping('AwsLbAccounts', aws_mappings.AWS_LB_ACCOUNTS)
    template.add_condition(
        USE_CLOUDMAP_CON_T,
        cfn_conditions.USE_CLOUDMAP_CON
    )
    template.add_condition(
        cfn_conditions.NOT_USE_CLOUDMAP_CON_T,
        cfn_conditions.NOT_USE_CLOUDMAP_CON
    )
    vpc = add_vpc_core(template, cidr_block)
    storage_subnets = add_storage_subnets(template, vpc[0], az_range, layers)
    public_subnets = add_public_subnets(template, vpc[0], az_range, layers, vpc[-1], single_nat)
    app_subnets = add_apps_subnets(template, vpc[0], az_range, layers, public_subnets[-1])
    add_template_outputs(
        template,
        vpc[0],
        storage_subnets[1],
        public_subnets[1],
        app_subnets[1],
    )
    add_vpc_cidrs_outputs(template, layers)
    add_cloudmap_support(template, vpc[0])
    validate_template(template.to_json(), 'vpc.json')
    with open('/tmp/vpc.json', 'w') as fd:
        fd.write(template.to_json())
    return template



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module to generate a cluster template standalone or import create_cluster_template
elsewhere for nested cluster stack
"""

import argparse
import sys

from boto3 import session

from ecs_composex.compute import create_cluster_template
from ecs_composex.common.aws import CURATED_AZS, BUCKET_NAME
from ecs_composex.ecs.ecs_params import CLUSTER_NAME_T
from ecs_composex.vpc.vpc_params import (
    VPC_ID_T, APP_SUBNETS_T,
    PUBLIC_SUBNETS_T
)


def root_parser():
    """
    Function to create the VPC specific arguments for argparse
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-f', '--docker-compose-file', required=False
    )
    parser.add_argument(
        '-o', '--output-file', required=True, help="Output file"
    )
    #  AWS SETTINGS
    parser.add_argument(
        '--region', required=False, default=session.Session().region_name,
        dest='AwsRegion',
        help="Specify the region you want to build for"
             "default use default region from config or environment vars"
    )
    parser.add_argument(
        '--az', dest='AwsAzs', action='append', required=False, default=CURATED_AZS,
        help="List AZs you want to deploy to specifically within the region"
    )
    parser.add_argument(
        '-b', '--bucket-name', type=str, required=False, default=BUCKET_NAME,
        help='Bucket name to upload the templates to', dest='BucketName'
    )
    # VPC SETTINGS
    parser.add_argument(
        '--create-vpc', required=False, default=False, action='store_true',
        help="Create a VPC for this deployment", dest='CreateVpc'
    )
    parser.add_argument(
        '--vpc-cidr', required=False, default='192.168.36.0/22', dest='VpcCidr',
        help="Specify the VPC CIDR if you use --create-vpc"
    )
    parser.add_argument(
        '--vpc-id', dest=VPC_ID_T, required=False, type=str,
        help='Specify VPC ID when not creating one'
    )
    parser.add_argument(
        '--public-subnets', required=False, dest=PUBLIC_SUBNETS_T, action='append',
        help="List of Subnet IDs to use for the cluster when not creating VPC"
    )
    parser.add_argument(
        '--app-subnets', required=False, dest=APP_SUBNETS_T, action='append',
        help="List of Subnet IDs to use for the cluster when not creating VPC"
    )
    parser.add_argument(
        '--storage-subnets', required=False, dest=APP_SUBNETS_T, action='append',
        help="List of Subnet IDs to use for the cluster when not creating VPC"
    )
    # CLUSTER SETTINGS
    parser.add_argument(
        '--create-cluster', required=False, default=False, action='store_true',
        help="Create an ECS Cluster for this deployment", dest='CreateCluster'
    )
    parser.add_argument(
        '--cluster-name', dest=CLUSTER_NAME_T, required=False
    )
    parser.add_argument(
        '--use-spot-fleet', required=False, default=False, action='store_true',
        dest='UseSpotFleet',
        help="Runs spotfleet for EC2. If used in combination "
             "of --use-fargate, it will create an additional SpotFleet"
    )
    return parser


def main():
    """Console script for ecs_composex."""
    parser = root_parser()
    parser.add_argument('_', nargs='*')
    args = parser.parse_args()

    template = create_cluster_template(**vars(args))
    with open(args.output_file, 'w') as tpl_fd:
        if args.output_file.endswith('.yml') or args.output_file.endswith('.yaml'):
            tpl_fd.write(template.to_yaml())
        else:
            tpl_fd.write(template.to_json())
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



# -*- coding: utf-8 -*-

"""Common Conditions across the templates"""

from ecs_composex.common import cfn_params
from troposphere import (
    Condition, Not,
    Ref, Equals, And
)

USE_STACK_NAME_CON_T = 'UseStackName'
USE_STACK_NAME_CON = Equals(
    Ref(cfn_params.ROOT_STACK_NAME),
    cfn_params.ROOT_STACK_NAME.Default
)

USE_CLOUDMAP_CON_T = 'UseCloudMapCondition'
USE_CLOUDMAP_CON = Equals(
    Ref(cfn_params.SERVICE_DISCOVERY_T),
    'True'
)

NOT_USE_CLOUDMAP_CON_T = 'NotUseCloudMapCondition'
NOT_USE_CLOUDMAP_CON = Not(Condition(USE_CLOUDMAP_CON_T))

USE_CFN_PARAMS_CON_T = 'UseCfnParametersValueCondition'
USE_CFN_PARAMS_CON = Equals(
    Ref(cfn_params.USE_CFN_PARAMS),
    cfn_params.USE_CFN_PARAMS.Default
)

NOT_USE_CFN_PARAMS_CON_T = f"Not{USE_CFN_PARAMS_CON_T}"
NOT_USE_CFN_PARAMS_CON = Not(Condition(USE_CFN_PARAMS_CON_T))

USE_CFN_EXPORTS_T = 'UseExportsCondition'
USE_CFN_EXPORTS = Equals(
    Ref(cfn_params.USE_CFN_EXPORTS),
    'True'
)

NOT_USE_CFN_EXPORTS_T = 'NotUseCfnExportsCondition'
NOT_USE_CFN_EXPORTS = Not(Condition(USE_CFN_EXPORTS_T))

USE_SSM_EXPORTS_T = 'UseSsmExportsCondition'
USE_SSM_EXPORTS = Equals(
    Ref(cfn_params.USE_SSM_EXPORTS),
    'True'
)

USE_CFN_AND_SSM_EXPORTS_T = 'UseCfnAndSsmCondition'
USE_CFN_AND_SSM_EXPORTS = And(
    Condition(USE_CFN_EXPORTS_T),
    Condition(USE_SSM_EXPORTS_T)
)

USE_SSM_ONLY_T = 'UseSsmOnlyCondition'
USE_SSM_ONLY = And(
    Condition(USE_SSM_EXPORTS_T),
    Condition(NOT_USE_CFN_EXPORTS_T)
)



# -*- coding: utf-8 -*-
#!/usr/bin/env python

import os
import yaml
import pytest
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from ecs_composex.common import build_template
from ecs_composex.ecs.ecs_iam import (
    add_service_roles,
    assign_x_resources_to_service
)

try:
    BUCKET = os.environ['KNOWN_BUCKET']
except KeyError:
    BUCKET = 'lambda-dev-eu-west-1'


HERE = os.path.dirname(os.path.abspath(__file__))

@pytest.fixture
def content():
    """
    Opens the file and passes the content around
    """
    with open(f"{HERE}/services_with_queues.yml", 'r') as fd:
        content = yaml.load(fd.read(), Loader=Loader)
    return content

@pytest.fixture
def config():
    """
    Config object
    """
    return {
        'BucketName': BUCKET,
        'EnvName': 'abcd',
        'Debug': True
    }


def test_ecs_roles(config):
    """
    Tests the creation of service roles
    """
    tmp_tpl = build_template('Tmp template')
    add_service_roles(tmp_tpl)


def test_ecs_roles_permissions(content, config):
    """
    Function to test
    """
    tmp_tpl = build_template('TMP Template')
    add_service_roles(tmp_tpl)
    assign_x_resources_to_service(content, 'app01', tmp_tpl, **config)



# -*- coding: utf-8 -*-
"""
Parameters bound to  ecs_composex.ecs
This is a crucial part as all the titles, maked `_T` are string which are then used the same way
across all imports, which gives consistency for CFN to use the same names,
which it heavily relies onto.

You can change the names *values* so you like so long as you keep it [a-zA-Z0-9]
"""

from troposphere import Parameter, Select, Split, Ref, ImportValue, Sub

from ecs_composex.common.cfn_params import ROOT_STACK_NAME_T
from ecs_composex.vpc.vpc_params import SG_ID_TYPE

LOG_GROUP_T = 'ServicesLogGroup'
SG_T = 'ServiceSecurityGroup'
NETWORK_MODE = 'awsvpc'
EXEC_ROLE_T = 'EcsExecutionRole'
TASK_ROLE_T = 'EcsTaskRole'
SERVICE_T = 'EcsServiceDefinition'
TASK_T = 'EcsTaskDefinition'
RES_KEY = 'services'

LAUNCH_TYPE_T = 'EcsLaunchType'
LAUNCH_TYPE = Parameter(
    LAUNCH_TYPE_T, Type='String',
    AllowedValues=['EC2', 'FARGATE'],
    Default='FARGATE'
)

IS_PUBLIC_T = 'ExposeServicePublicly'
IS_PUBLIC = Parameter(
    IS_PUBLIC_T,
    AllowedValues=['True', 'False'],
    Type='String'
)

TASK_CPU_COUNT_T = 'TaskCpuCount'
TASK_CPU_COUNT = Parameter(
    TASK_CPU_COUNT_T,
    Type='Number',
    Default=1024
)
MEMORY_ALLOC_T = 'ContainerMemoryAllocation'
MEMORY_RES_T = 'ContainerMemoryReservation'

MEMORY_ALLOC = Parameter(MEMORY_ALLOC_T, Type='Number', Default=512)
MEMORY_RES = Parameter(MEMORY_RES_T, Type='Number', Default=0)

CLUSTER_NAME_T = 'EcsClusterName'
CLUSTER_NAME = Parameter(
    CLUSTER_NAME_T, Type='String',
    AllowedPattern=r'[a-zA-Z0-9-]+',
    Default='default'
)

SERVICE_NAME_T = 'MicroServiceName'
SERVICE_NAME = Parameter(
    SERVICE_NAME_T, Type='String',
    AllowedPattern=r'[a-zA-Z0-9-]+'
)

SERVICE_IMAGE_T = 'MicroserviceImage'
SERVICE_IMAGE = Parameter(SERVICE_IMAGE_T, Type='String')

SERVICE_COUNT_T = 'MicroservicesCount'
SERVICE_COUNT = Parameter(
    SERVICE_COUNT_T,
    Type='Number',
    MinValue=0,
    Default=0
)

ELB_GRACE_PERIOD_T = 'ElbGracePeriod'
ELB_GRACE_PERIOD = Parameter(
    ELB_GRACE_PERIOD_T,
    Type='Number',
    MinValue=0,
    Default=90,
    MaxValue=300
)

ECS_CONTROLLER_T = 'EcsServiceDeploymentController'
ECS_CONTROLLER = Parameter(
    ECS_CONTROLLER_T,
    Type='String',
    AllowedValues=[
        'ECS',
        'CODE_DEPLOY',
        'EXTERNAL'
    ],
    Default='ECS'
)

LOG_GROUP = Parameter(
    f"Cluster{LOG_GROUP_T}",
    Type='String'
)

FARGATE_CPU_RAM_CONFIG_T = 'FargateCpuRamConfiguration'
FARGATE_CPU_RAM_CONFIG = Parameter(
    FARGATE_CPU_RAM_CONFIG_T,
    Type='String',
    AllowedValues=[
        '256!512',
        '256!1024',
        '256!2048',
        '512!1024',
        '512!2048',
        '512!3072',
        '512!4096',
        '1024!2048',
        '1024!3072',
        '1024!4096',
        '1024!5120',
        '1024!6144',
        '1024!7168',
        '1024!8192',
        '2048!2048',
        '2048!3072',
        '2048!4096',
        '2048!5120',
        '2048!6144',
        '2048!7168',
        '2048!8192',
        '2048!9216',
        '2048!10240',
        '2048!11264',
        '2048!12288',
        '2048!13312',
        '2048!14336',
        '2048!15360',
        '2048!16384',
        '4096!8192',
        '4096!9216',
        '4096!10240',
        '4096!11264',
        '4096!12288',
        '4096!13312',
        '4096!14336',
        '4096!15360',
        '4096!16384',
        '4096!17408',
        '4096!18432',
        '4096!19456',
        '4096!20480',
        '4096!21504',
        '4096!22528',
        '4096!23552',
        '4096!24576',
        '4096!25600',
        '4096!26624',
        '4096!27648',
        '4096!28672',
        '4096!29696',
        '4096!30720'
    ],
    Default='512!1024'
)

FARGATE_CPU = Select(0, Split('!', Ref(FARGATE_CPU_RAM_CONFIG)))
FARGATE_RAM = Select(1, Split('!', Ref(FARGATE_CPU_RAM_CONFIG)))

CLUSTER_SG_ID_T = 'ClusterWideSGId'
CLUSTER_SG_ID = Parameter(
    CLUSTER_SG_ID_T,
    Type=SG_ID_TYPE,
    Default='<none>'
)

SERVICE_GROUP_ID_T = 'ServiceGroupId'
SERVICE_GROUP_ID = Parameter(
    SERVICE_GROUP_ID_T,
    Type=SG_ID_TYPE,
    Default='<none>'
)


def get_import_service_group_id(remote_service_name):
    """
    Function to return the ImportValue(Sub()) for given service name
    """
    return ImportValue(Sub(f"${{{ROOT_STACK_NAME_T}}}-{remote_service_name}-{SERVICE_GROUP_ID_T}"))



# -*- coding: utf-8 -*-
"""
Set of functions to generate permissions to access queues
based on pre-defined SQS policies for consumers
"""

from troposphere import Sub, ImportValue, If
from troposphere.iam import Policy as IamPolicy
from troposphere.ecs import Environment
from ecs_composex.sqs import sqs_params
from ecs_composex.common import LOG
from ecs_composex.common.cfn_params import ROOT_STACK_NAME_T
from ecs_composex.common.cfn_conditions import USE_SSM_ONLY_T
from ecs_composex.common import KEYISSET
from ecs_composex.sqs import SQS_SSM_PREFIX

QUEUES_ACCESS_TYPES = {
    'RWMessages': {
        "NotAction": [
            "sqs:TagQueue",
            "sqs:RemovePermission",
            "sqs:AddPermission",
            "sqs:UntagQueue",
            "sqs:PurgeQueue",
            "sqs:DeleteQueue",
            "sqs:CreateQueue",
            "sqs:SetQueueAttributes"
        ],
        "Effect": "Allow"
    },
    'RWPermissions': {
        "NotAction": [
            "sqs:RemovePermission",
            "sqs:AddPermission",
            "sqs:PurgeQueue",
            "sqs:SetQueueAttributes"
        ],
        "Effect": "Allow"
    },
    'RO': {
        "NotAction": [
            "sqs:TagQueue",
            "sqs:RemovePermission",
            "sqs:AddPermission",
            "sqs:UntagQueue",
            "sqs:PurgeQueue",
            "sqs:Delete*",
            "sqs:Create*",
            "sqs:Set*"
        ],
        "Effect": "Allow"
    }
}


def generate_queue_strings(queue_name):
    """
    Function to generate the SSM and CFN import/export strings
    Returns the import in a tuple

    :param queue_name: name of the queue as defined in ComposeX File
    :type queue_name:

    :returns: tuple(ssm_export, cfn_export)
    :rtype: tuple(Sub, ImportValue)
    """
    ssm_string = f"/${{{ROOT_STACK_NAME_T}}}{SQS_SSM_PREFIX}{queue_name}"
    ssm_export = Sub(r'{{resolve:ssm:%s:1}}' % (ssm_string))
    cfn_string = f"${{{ROOT_STACK_NAME_T}}}-{queue_name}-{sqs_params.SQS_ARN_T}"
    cfn_import = ImportValue(Sub(cfn_string))
    return (ssm_export, cfn_import)


def generate_sqs_permissions(queue_name, resource, **kwargs):
    """
    Generates an IAM policy for each access type and returns a dictionnary of these.

    :params queue_name: String of the name of the queue as defined in Docker compose
    :type queue_name: str
    :param resource: The Troposphere resource
    :type resource: dict
    :param config: Dictionnary containing the configuration for rendering the permissions
    :type config: dict

    :returns: Dictionnary of IAM policies (PolicyType)
    :rtype: dict
    """

    export_strings = generate_queue_strings(queue_name)
    queue_policies = {}
    services = resource['Services']
    for a_type in QUEUES_ACCESS_TYPES:
        clean_policy = {'Version': '2012-10-17', 'Statement': []}
        LOG.debug(a_type)
        policy_doc = QUEUES_ACCESS_TYPES[a_type].copy()
        policy_doc['Resource'] = If(
            USE_SSM_ONLY_T,
            export_strings[0],
            export_strings[1]
        )
        clean_policy['Statement'].append(policy_doc)
        queue_policies[a_type] = {
            'Services': [],
            'Policy': None
        }
        queue_policies[a_type]['Policy'] = IamPolicy(
            PolicyName=Sub(f"AccessToSqsQueue{queue_name}In${{{ROOT_STACK_NAME_T}}}"),
            PolicyDocument=clean_policy
        )
        for service in services:
            if service['access'] == a_type:
                queue_policies[a_type]['Services'].append(service['name'])
    return queue_policies


def generate_sqs_envvars(queue_name, resource, **kwargs):
    """
    Function to generate environment variables that can be added to a container definition
    shall the service need to know about the Queue
    """
    env_names = []
    export_strings = generate_queue_strings(queue_name)
    if KEYISSET('Settings', resource) and KEYISSET('EnvNames', resource['Settings']):
        for env_name in resource['Settings']['EnvNames']:
            env_names.append(Environment(
                    Name=env_name,
                    Value=If(
                            USE_SSM_ONLY_T,
                            export_strings[0],
                            export_strings[1]
                        )
                )
            )
        if queue_name not in resource['Settings']['EnvNames']:
            env_names.append(Environment(
                    Name=queue_name,
                    Value=If(
                            USE_SSM_ONLY_T,
                            export_strings[0],
                            export_strings[1]
                        )
                )
            )
    return env_names



# -*- coding: utf-8 -*-
"""Module to handle AWS SQS CFN Templates creation"""

import os
import boto3

from ecs_composex.common import LOG
from ecs_composex.common import (
    validate_input,
    validate_kwargs,
    load_composex_file
)
from ecs_composex.common.templates import validate_template
from ecs_composex.sqs.sqs_template import generate_sqs_root_template

RES_KEY = f"x-{os.path.basename(os.path.dirname(os.path.abspath(__file__)))}"
SQS_SSM_PREFIX = f"/{RES_KEY}/"


def create_sqs_template(session=None, **kwargs):
    """
    Creates the CFN Troposphere template

    :param session: boto3 session to override default
    :type session: boto3.session.Session

    :return: sqs_tpl
    :rtype: troposphere.Template
    """
    content = load_composex_file(kwargs['ComposeXFile'])
    validate_input(content, RES_KEY)
    validate_kwargs(['BucketName'], kwargs)

    if session is None:
        session = boto3.session.Session()
    sqs_tpl = generate_sqs_root_template(compose_content=content, session=session, **kwargs)

    validate_template(sqs_tpl.to_json(), 'sqs_root')
    LOG.debug("Template for SQS validated by CFN.")
    return sqs_tpl



# -*- coding: utf-8 -*-

"""
Functions to add the three VPC layer type subnets:

* Storage
* Public
* App

RTB -> Route Table

Storage subnet type : All subnets use the same RTB, no route to 0.0.0.0/0
Public subnet type: All subnets use the same RTB, route to 0.0.0.0/0 via InternetGateway
App subnet type: Each subnet has its own RTB, each RTB points to a different NAT Gateway in its
respective AZ

"""

from string import ascii_lowercase as alpha
from troposphere.ec2 import (
    Subnet, SubnetRouteTableAssociation,
    NatGateway, EIP, Route, RouteTable
)
from troposphere import (
    GetAtt, Tags, Ref, Sub, If
)

from ecs_composex.common.cfn_params import ROOT_STACK_NAME_T
from ecs_composex.common.cfn_conditions import USE_STACK_NAME_CON_T


def add_storage_subnets(template, vpc, az_range, layers):
    """
    Function to add storage subnets inside the VPC

    :param template: VPC Template()
    :param vpc: Vpc() for Ref()
    :param az_range: range for iteration over select AZs

    :returns: tuple() list of rtb, list of subnets
    """
    rtb = RouteTable(
        'StorageRtb',
        template=template,
        VpcId=Ref(vpc),
        Tags=Tags(Name='StorageRtb') + Tags({'vpc:usage': 'storage'})
    )

    subnets = []
    for count, subnet_cidr in zip(az_range, layers['stor']):
        subnet = Subnet(
            f'StorageSubnet{alpha[count].upper()}',
            template=template,
            CidrBlock=subnet_cidr,
            VpcId=Ref(vpc),
            AvailabilityZone=Sub(f'${{AWS::Region}}{alpha[count]}'),
            Tags=Tags(
                Name=If(
                    USE_STACK_NAME_CON_T,
                    Sub(f'${{AWS::StackName}}-Storage-{alpha[count]}'),
                    Sub(f'${{{ROOT_STACK_NAME_T}}}-Storage-{alpha[count]}')
                ),
            ) + Tags({'vpc:usage': 'storage'})
        )
        SubnetRouteTableAssociation(
            f'StorageSubnetAssoc{alpha[count].upper()}',
            template=template,
            SubnetId=Ref(subnet),
            RouteTableId=Ref(rtb)
        )
        subnets.append(subnet)
    return ([rtb], subnets)


def add_public_subnets(template, vpc, az_range, layers, igw, single_nat):
    """
    Function to add public subnets for the VPC

    :param template: VPC Template()
    :param vpc: Vpc() for Ref()
    :param az_range: range for iteration over select AZs

    :returns: tuple() list of rtb, list of subnets, list of nats
    """
    rtb = RouteTable(
        'PublicRtb',
        template=template,
        VpcId=Ref(vpc),
        Tags=Tags(Name='PublicRtb') + Tags({'vpc:usage': 'public'})
    )
    Route(
        'PublicDefaultRoute',
        template=template,
        GatewayId=Ref(igw),
        RouteTableId=Ref(rtb),
        DestinationCidrBlock='0.0.0.0/0'
    )
    subnets = []
    nats = []
    for count, subnet_cidr in zip(az_range, layers['pub']):
        subnet = Subnet(
            f'PublicSubnet{alpha[count].upper()}',
            template=template,
            CidrBlock=subnet_cidr,
            VpcId=Ref(vpc),
            AvailabilityZone=Sub(f'${{AWS::Region}}{alpha[count]}'),
            MapPublicIpOnLaunch=True,
            Tags=Tags(
                Name=If(
                    USE_STACK_NAME_CON_T,
                    Sub(f'${{AWS::StackName}}-Public-{alpha[count]}'),
                    Sub(f'${{{ROOT_STACK_NAME_T}}}-Public-{alpha[count]}')
                ),
            ) + Tags({'30-automation:usage': 'public'})
        )
        if (single_nat and not nats) or not single_nat:
            eip = EIP(
                f"NatGatewayEip{alpha[count].upper()}",
                template=template,
                Domain='vpc'
            )
            nat = NatGateway(
                f"NatGatewayAz{alpha[count].upper()}",
                template=template,
                AllocationId=GetAtt(eip, 'AllocationId'),
                SubnetId=Ref(subnet)
                )
            nats.append(nat)
        SubnetRouteTableAssociation(
            f'PublicSubnetsRtbAssoc{alpha[count].upper()}',
            template=template,
            RouteTableId=Ref(rtb),
            SubnetId=Ref(subnet)
        )
        subnets.append(subnet)
    return ([rtb], subnets, nats)


def add_apps_subnets(template, vpc, az_range, layers, nats):
    """
    Function to add application/hosts subnets to the VPC

    :param template: VPC Template()
    :param vpc: Vpc() for Ref()
    :param az_range: range for iteration over select AZs
    :param nats: list of NatGateway()

    :returns: tuple() list of rtb, list of subnets
    """
    subnets = []
    rtbs = []
    if len(nats) < len(az_range):
        primary_nat = nats[0]
        nats = []
        for az in az_range:
            nats.append(primary_nat)
    for count, subnet_cidr, nat in zip(az_range, layers['app'], nats):
        suffix = alpha[count].upper()
        subnet = Subnet(
            f'AppSubnet{suffix}',
            template=template,
            CidrBlock=subnet_cidr,
            VpcId=Ref(vpc),
            AvailabilityZone=Sub(f'${{AWS::Region}}{alpha[count]}'),
            Tags=Tags(
                Name=If(
                    USE_STACK_NAME_CON_T,
                    Sub(f'${{AWS::StackName}}-App-{alpha[count]}'),
                    Sub(f'${{{ROOT_STACK_NAME_T}}}-App-{alpha[count]}')
                ),
            ) + Tags({'vpc:usage': 'application'})
        )
        rtb = RouteTable(
            f'AppRtb{alpha[count].upper()}',
            template=template,
            VpcId=Ref(vpc),
            Tags=Tags(Name=f'AppRtb{alpha[count].upper()}')
        )
        Route(
            f'AppRoute{alpha[count].upper()}',
            template=template,
            NatGatewayId=Ref(nat),
            RouteTableId=Ref(rtb),
            DestinationCidrBlock='0.0.0.0/0'
        )
        SubnetRouteTableAssociation(
            f'SubnetRtbAssoc{alpha[count].upper()}',
            template=template,
            RouteTableId=Ref(rtb),
            SubnetId=Ref(subnet)
        )
        rtbs.append(rtb)
        subnets.append(subnet)
    return rtbs, subnets



# -*- coding: utf-8 -*-
"""
Functions to build the ECS Service Definition
"""

from troposphere import (
    Tags, GetAtt,
    Ref, If, Join
)
from troposphere.ecs import (
    Service,
    PlacementStrategy,
    AwsvpcConfiguration,
    NetworkConfiguration,
    DeploymentController
)

from ecs_composex.common import LOG, cfn_conditions
from ecs_composex.common import (
    build_template, cfn_params
)
from ecs_composex.common.cfn_params import (
    ROOT_STACK_NAME_T
)
from ecs_composex.common.outputs import formatted_outputs
from ecs_composex.common.templates import upload_template
from ecs_composex.ecs import ecs_conditions
from ecs_composex.ecs import ecs_params
from ecs_composex.ecs.ecs_iam import (
    add_service_roles,
    assign_x_resources_to_service
)
from ecs_composex.ecs.ecs_loadbalancing import define_grace_period
from ecs_composex.ecs.ecs_networking import (
    define_service_network_config,
    compile_network_settings
)
from ecs_composex.ecs.ecs_networking_ingress import define_service_to_service_ingress
from ecs_composex.ecs.ecs_task import (
    add_task_defnition
)
from ecs_composex.vpc import vpc_params

STATIC = 0


def define_placement_strategies():
    """
    Function to generate placement strategies
    :return: list of placement strategies
    :rtype: list
    """
    return [
        PlacementStrategy(
            Field='instanceId',
            Type='spread'
        ),
        PlacementStrategy(
            Field='attribute:ecs.availability-zone',
            Type='spread'
        )
    ]


def generate_service_definition(template, network_settings, security_groups, **kwargs):
    """Function to generate the Service definition

    :param template: service template
    :type template: troposphere.Template
    :param network_settings: network settings as defined in compile_network_settings
    :type network_settings: dict
    :param security_groups: list of security groups for the service to use
    :type security_groups: list
    :param kwargs: extra settings that the Service() object can add to
    :type kwargs: dict
    """

    service_sgs = [Ref(sg) for sg in security_groups]
    Service(
        ecs_params.SERVICE_T,
        template=template,
        Cluster=Ref(ecs_params.CLUSTER_NAME),
        DeploymentController=DeploymentController(Type=Ref(ecs_params.ECS_CONTROLLER)),
        EnableECSManagedTags=True,
        DesiredCount=If(
            ecs_conditions.SERVICE_COUNT_ZERO_AND_FARGATE_CON_T,
            1,
            If(
                ecs_conditions.USE_FARGATE_CON_T,
                Ref(ecs_params.SERVICE_COUNT),
                If(
                    ecs_conditions.SERVICE_COUNT_ZERO_CON_T,
                    Ref('AWS::NoValue'),
                    Ref(ecs_params.SERVICE_COUNT)
                )
            )
        ),
        SchedulingStrategy=If(
            ecs_conditions.USE_FARGATE_CON_T,
            'REPLICA',
            If(
                ecs_conditions.SERVICE_COUNT_ZERO_AND_FARGATE_CON_T,
                'REPLICA',
                'DAEMON'
            )
        ),
        HealthCheckGracePeriodSeconds=define_grace_period(template, network_settings),
        PlacementStrategies=If(
            ecs_conditions.USE_FARGATE_CON_T,
            Ref('AWS::NoValue'),
            define_placement_strategies()
        ),
        NetworkConfiguration=NetworkConfiguration(
            AwsvpcConfiguration=AwsvpcConfiguration(
                Subnets=Ref(vpc_params.APP_SUBNETS),
                SecurityGroups=service_sgs
            )
        ),
        TaskDefinition=Ref(ecs_params.TASK_T),
        LaunchType=Ref(ecs_params.LAUNCH_TYPE),
        Tags=Tags(
            {
                'Name': Ref(ecs_params.SERVICE_NAME),
                'StackName': Ref('AWS::StackName')
            }
        ),
        PropagateTags='SERVICE',
        **kwargs
    )


def initialize_service_template(service_name):
    """Function to initialize the base template for ECS Services with all
    parameters and conditions necessary for CFN to work properly

    :param service_name: Name of the service as defined in ComposeX File
    :type service_name: str

    :return: service_template
    :rtype: troposphere.Template
    """
    service_tpl = build_template(
        f"Template for {service_name}",
        [
            ecs_params.CLUSTER_NAME,
            ecs_params.LAUNCH_TYPE,
            cfn_params.SERVICE_DISCOVERY,
            ecs_params.ECS_CONTROLLER,
            ecs_params.SERVICE_COUNT,
            ecs_params.CLUSTER_SG_ID,
            vpc_params.VPC_ID,
            vpc_params.APP_SUBNETS,
            vpc_params.PUBLIC_SUBNETS,
            ecs_params.LOG_GROUP
        ]
    )
    service_tpl.add_condition(
        ecs_conditions.MEM_RES_IS_MEM_ALLOC_CON_T,
        ecs_conditions.MEM_RES_IS_MEM_ALLOC_CON
    )
    service_tpl.add_condition(
        cfn_conditions.USE_CLOUDMAP_CON_T,
        cfn_conditions.USE_CLOUDMAP_CON
    )
    service_tpl.add_condition(
        ecs_conditions.SERVICE_COUNT_ZERO_CON_T,
        ecs_conditions.SERVICE_COUNT_ZERO_CON
    )
    service_tpl.add_condition(
        ecs_conditions.SERVICE_COUNT_ZERO_AND_FARGATE_CON_T,
        ecs_conditions.SERVICE_COUNT_ZERO_AND_FARGATE_CON
    )
    return service_tpl


def generate_service_template_outputs(template, service_name):
    """
    Function to generate the Service template outputs

    :param template: service template
    :type template: troposphere.Template
    :param service_name: name of the service as defined in Docker ComposeX file
    """
    template.add_output(formatted_outputs([
        {
            ecs_params.SERVICE_GROUP_ID_T: GetAtt(ecs_params.SG_T, 'GroupId')
        }
    ], export=True, prefix=f"${{{ROOT_STACK_NAME_T}}}-{service_name}"))


def generate_service_template(compose_content, service_name, service, session=None, **kwargs):
    """
    Function to generate single service template based on its definition in
    the Compose file.

    :param compose_content: Docker/ComposeX Content
    :type compose_content: dict
    :param service_name: Name of the service as defined in ComposeX
    :type service_name: str
    :param service: service dict as defined in ComposeX
    :type service: dict
    :param session: boto3 session for override
    :type session: boto3.session.Session

    :returns: service template URL, service specific parameters, stack dependencies
    :rtype: tuple
    """
    network_settings = compile_network_settings(
        compose_content, service, service_name
    )
    service_tpl = initialize_service_template(service_name)
    parameters = {
        vpc_params.VPC_ID_T: Ref(vpc_params.VPC_ID),
        vpc_params.APP_SUBNETS_T: Join(',', Ref(vpc_params.APP_SUBNETS)),
        vpc_params.PUBLIC_SUBNETS_T: Join(',', Ref(vpc_params.PUBLIC_SUBNETS)),
        ecs_params.CLUSTER_NAME_T: Ref(ecs_params.CLUSTER_NAME),
        ecs_params.LOG_GROUP.title: Ref(ecs_params.LOG_GROUP_T)
    }
    add_service_roles(service_tpl)
    parameters.update(add_task_defnition(
        service_tpl, service_name, service, network_settings
    ))
    assign_x_resources_to_service(
        compose_content, service_name,
        service_tpl, **kwargs
    )
    service_sgs = [ecs_params.SG_T, ecs_params.CLUSTER_SG_ID]
    service_network_config = define_service_network_config(
        service_tpl, service_name, network_settings, **kwargs
    )
    generate_service_definition(
        service_tpl, network_settings, service_sgs,
        **service_network_config[0]
    )
    services_dependencies = define_service_to_service_ingress(
        compose_content, service_tpl, service_name, service
    )
    stack_dependencies = []
    if isinstance(service_network_config[-1], list):
        stack_dependencies += service_network_config[-1]
    if isinstance(services_dependencies, list):
        stack_dependencies += services_dependencies
    generate_service_template_outputs(service_tpl, service_name)
    service_tpl_url = upload_template(
        service_tpl.to_json(),
        kwargs['BucketName'],
        f"{service_name}.json",
        session=session
    )
    LOG.debug(service_tpl_url)
    return service_tpl_url, parameters, stack_dependencies



# -*- coding: utf-8 -*-
"""Main module that allows expansion of ECS ComposeX to add new packages to
support more AWS resources, such as SQS, SNS, RDS etc.

Every package must have a module named {resource_key}_perms so that functions
that return policies and other attributes can find these and execute so the
services can add the policies and environment variable to the task definition
and IAM task role.
"""

import re
from importlib import import_module
from ecs_composex.common import KEYISSET, validate_resource_title, LOG

RES_REGX = re.compile(r"(^([x-]+))")


def get_composex_globals(compose_content):
    """Parses configs and looks for globals
    :param compose_content: the docker composeX content
    :type compose_content: dict

    :return: docker compose globals
    :rtype: dict
    """
    if KEYISSET('configs', compose_content) and KEYISSET('composex', compose_content['configs']):
        return compose_content['configs']['composex']
    return {}


def get_mod_function(module_name, function_name):
    """Gets the permissions function from module

    :param module_name: the name of the module in ecs_composex to find and try to import
    :type module_name: str
    :param function_name: name of the function to try to get
    :type function_name: str

    :return: function, if found, from the module
    :rtype: function
    """
    composex_module_name = f"ecs_composex.{module_name}"
    LOG.debug(composex_module_name)
    res_module = None
    try:
        res_module = import_module(composex_module_name)
        LOG.debug(res_module)
        try:
            function = getattr(res_module, function_name)
            return function
        except AttributeError:
            LOG.info(f"No {function_name} function found - skipping")
    except ImportError as error:
        LOG.error(
            f'Failure to process the module {composex_module_name}'
        )
        LOG.error(error)
    return res_module


def generate_x_resources_policies(resources, resource_type, function, **kwargs):
    """Function to create the policies for the resources of a given type

    :param resources: resources to go over from docker ComposeX file
    :type resources: dict
    :param resource_type: resource type, ie. sqs
    :type resource_type: str
    :param function: the function to call to get IAM policies to add to service role
    :type function: function pointer
    :param kwargs: optional arguments
    :type kwargs: dict

    :return: policies for each resource of given resource type to use for roles
    :rtype: dict
    """
    mod_policies = {}
    for resource_name in resources:
        assert validate_resource_title(resource_name, resource_type)
        resource = resources[resource_name]

        if KEYISSET('Services', resource):
            mod_policies[resource_name] = function(
                resource_name, resource, **kwargs
            )
    return mod_policies


def generate_x_resources_envvars(resources, resource_type, function, **kwargs):
    """Function to create the env vars for the resources of given type

    :param resources: resources to go over for generating envvars
    :type resources: dict
    :param resource_type: resource type, i.e., sqs
    :type resource_type: str
    :param function: the function to use to fetch the information needed
    :type function: function pointer
    :param kwargs: optional arguments
    :type kwargs: dict

    :return: envvars for resources of the given resource type to be used by microservices for environment variables
    :rtype: dict
    """
    mod_envvars = {}
    for resource_name in resources:
        assert validate_resource_title(resource_name, resource_type)
        resource = resources[resource_name]

        if KEYISSET('Services', resource):
            mod_envvars[resource_name] = function(
                resource_name, resource, **kwargs
            )
    return mod_envvars


def generate_x_resource_configs(content, **kwargs):
    """
    Function that evaluates only the x- sections of the Compose file
    and generates calls the init function for each.

    :param content: docker ComposeX file content
    :param kwargs: settings for building X related resources

    :return: resource_configs
    :rtype: dict
    """
    resource_configs = {}
    options = get_composex_globals(content)
    kwargs.update(options)
    for resource_type in content.keys():
        res_name = RES_REGX.sub('', resource_type)
        resources = content[resource_type]
        if (resource_type.startswith('x-')
                and not (resource_type == 'x-rds' or resource_type == 'x-cluster')
                and content[resource_type]):
            resource_configs[resource_type] = {}
            module_name = f"{res_name}.{res_name}_perms"
            perms_function_name = f"generate_{res_name}_permissions"
            vars_function_name = f"generate_{res_name}_envvars"
            perms_function = get_mod_function(module_name, perms_function_name)
            vars_function = get_mod_function(module_name, vars_function_name)
            LOG.debug(perms_function)
            LOG.debug(vars_function)
            if perms_function:
                resource_configs[resource_type]['permissions'] = generate_x_resources_policies(
                    resources,
                    resource_type,
                    perms_function,
                    **kwargs
                )
            if vars_function:
                resource_configs[resource_type]['envvars'] = generate_x_resources_envvars(
                    resources,
                    resource_type,
                    vars_function,
                    **kwargs
                )
    return resource_configs
